import dataclasses
import typing
import time

from weave.graph import Node
from ..api import op, weave_class, mutation
from .. import weave_types as types
from .. import errors
from .. import storage
from .. import registry_mem
from .. import weave_internal
from .. import trace
from .. import ref_base
from .. import uris
from .. import graph
from .. import artifact_local
from .. import compile
from .. import runs
from .. import artifact_fs
from .. import artifact_wandb
from .. import object_context


@weave_class(weave_type=types.RefType)
class RefNodeMethods:
    @op()
    def type(self) -> types.Type:
        self_ref = storage._get_ref(self)
        if self_ref is None:
            raise errors.WeaveInternalError("Cannot get type of ref resolving to None")
        return self_ref.type

    @op(output_type=lambda input_type: input_type["self"].object_type)
    def get(self):
        return storage.deref(self)

    # Tim: I commented this out as part of Weaveflow merge. There is a duplicate
    # definition below and i believe that will take precedence. Keeping here for
    # now in case we need to revert.
    # @op()
    # def __eq__(self, other: str) -> bool:
    #     return str(self) == uris.WeaveURI.parse(other).to_ref().uri

    @op()
    def __eq__(self, other: ref_base.Ref) -> bool:
        # Have to get the ref because the engine already dereffed
        other_ref = storage._get_ref(other)
        return str(self) == str(other_ref)


@op(
    input_type={
        "self": types.FilesystemArtifactRefType(types.Any()),
    },
)
def created_by(self) -> typing.Optional[runs.Run]:
    # TODO: engine derefences blindly before passing in, but we expect ref! Hack
    # here by just getting ._ref
    return trace.get_obj_creator(self._ref)


# Hmm... This returns the same obj, not a ref anymore
# TODO: is this what we want?
@op(
    name="save",
    output_type=lambda input_types: input_types["obj"],
)
def save(obj: typing.Any, name: typing.Optional[str]):
    ref = storage.save(obj, name=name)
    return ref.obj


# TODO: this is not used yet. We might want to call it save.
# it completes a transaction
@mutation
def finish(obj: graph.Node[typing.Any], name: typing.Optional[str] = None) -> None:
    nodes = graph.linearize(obj)
    if nodes is None:
        raise errors.WeaveInternalError("save must be called on a linear graph")
    node0 = nodes[0]
    if node0.from_op.name != "get":
        raise errors.WeaveInternalError("save must be called on a get node")
    target_uri = node0.from_op.inputs["uri"].val
    obj_ctx = object_context.get_object_context()
    if obj_ctx is None:
        raise errors.WeaveInternalError("save must be called inside a mutation")
    obj_ctx.finish_mutation(target_uri)


# TODO: This should not be a separate op.
# save above should return a ref. But some tests depend on it returning an object,
# which I didn't bother fixing yet.
@op(
    name="save_to_ref",
    output_type=types.RefType(),
    hidden=True,
)
def save_to_ref(obj: typing.Any, name: typing.Optional[str]):
    ref = storage.save(obj, name=name)
    return ref


@op(
    name="save_to_uri",
    output_type=types.RefType(),
    hidden=True,
)
def save_to_uri(obj: typing.Any, name: typing.Optional[str]):
    ref = storage.save(obj, name=name)
    return str(ref.uri)


def usedby_output_type(op_name: str) -> types.Type:
    op_def = registry_mem.memory_registry.get_op(op_name)

    if not callable(op_def.output_type):
        ot = op_def.output_type
    else:
        ot = op_def.concrete_output_type
        if isinstance(ot, types.UnknownType):
            # We can certainly fix this. But this would be a really cool case for
            # type variables. If this op outputs an output type with variables,
            # we could include them in the returned type here.
            # TODO: Fix
            raise errors.WeaveInternalError(
                "asking for used_by of op with callable output type not yet supported"
            )

    return types.List(types.RunType(inputs=op_def.input_type.weave_type(), output=ot))


def used_by_callable_output_type(input_type) -> types.Type:
    if not isinstance(input_type["op_name"], types.Const):
        return types.List(types.RunType())
    op_name = input_type["op_name"].val
    return usedby_output_type(op_name)


@op(render_info={"type": "function"}, hidden=True)
def usedby_refine_output_type(obj: typing.Any, op_name: str) -> types.Type:
    return usedby_output_type(op_name)


@op(
    pure=False,
    output_type=used_by_callable_output_type,
    refine_output_type=usedby_refine_output_type,
    hidden=True,
)
def used_by(obj: typing.Any, op_name: str):
    ref = storage.get_ref(obj)
    if ref is None:
        return []
    res = trace.used_by(ref, op_name)
    return res


@op()
def call_op(name: str) -> Node[types.Any]:
    return weave_internal.make_output_node(types.Any(), name, {})


@op(hidden=True)
def objects_refine_output_type(
    of_type: types.Type, alias: str, timestamp: int
) -> types.Type:
    return types.List(types.RefType(of_type))


def objects_output_type(input_type):
    if not isinstance(input_type["of_type"], types.Const):
        ref_obj_type = types.TypeType()
    else:
        ref_obj_type = input_type["of_type"].val

    return types.List(types.RefType(ref_obj_type))


@op(
    render_info={"type": "function"},
    output_type=objects_output_type,
    refine_output_type=objects_refine_output_type,
    # Not impure now that we have the cache-busting timestamp arg
    # pure=False,
)
# Timestamp is a way for us to force cache this, but also
# be able to bust the cache.
# TODO: Think about this pattern
def objects(of_type: types.Type, alias: str, timestamp: int):
    return storage.objects(of_type, alias)


# Return all local artifacts
# TODO: this is a placeholder used for loading dashboards. We won't really want to load
# all of them!
@op(render_info={"type": "function"})
def local_artifacts() -> list[artifact_local.LocalArtifact]:
    return storage.local_artifacts()


# TODO: used_by, created_by, and change them to use filter instead
#    to get filter output type right, I'll have to implement using refineOutputType
#    for now. We can optimize by moving this to the type system a little later.
# Then make a cool UI


def op_get_return_type(uri):
    ref = ref_base.Ref.from_str(uri)
    try:
        return ref.type
    except errors.WeaveArtifactVersionNotFound:
        return types.NoneType()


def get_const_val(list_type_or_node):
    if isinstance(list_type_or_node, Node):
        return get_const_val(list_type_or_node.type)
    elif isinstance(list_type_or_node, types.Const):
        return list_type_or_node.val
    else:
        raise errors.WeaveExpectedConstError("Expected a const value passed to `get`")


def op_get_return_type_from_inputs(inputs):
    return op_get_return_type(get_const_val(inputs["uri"]))


@op(
    name="getReturnType",
    input_type={"uri": types.String()},
    output_type=types.TypeType(),
    hidden=True,
    pure=False,
)
def get_returntype(uri):
    return op_get_return_type(uri)


def _save(
    get_uri,
    obj,
    root_args,
    make_new_type: typing.Callable[[types.Type], types.Type],
    mutation_record: object_context.MutationRecord,
):
    source_uri = uris.WeaveURI.parse(get_uri)

    target_branch = root_args.get("branch")
    if target_branch is None:
        target_branch = source_uri.version

        # If we are mutating a remote artifact, we want to make a
        # new branch with `user-` prefix. Still not convinced this is certainly
        # needed, but keeping in line with the client-enforced behavior.
        if isinstance(source_uri, artifact_wandb.WeaveWBArtifactURI):
            target_branch = "user-" + target_branch
    target_uri = artifact_local.WeaveLocalArtifactURI(
        source_uri.name, target_branch, source_uri.path  # type: ignore
    )

    ctx = object_context.get_object_context()
    if ctx:
        ctx.add_mutation(
            str(target_uri), str(source_uri), obj, make_new_type, mutation_record
        )
    else:
        # Save immediately
        with object_context.object_context() as ctx:
            ctx.add_mutation(
                str(target_uri), str(source_uri), obj, make_new_type, mutation_record
            )
    return str(target_uri)


@dataclasses.dataclass
class _MergeSpec:
    to_uri: typing.Union[
        artifact_wandb.WeaveWBArtifactURI, artifact_local.WeaveLocalArtifactURI
    ]
    head_ref: typing.Union[
        artifact_local.LocalArtifactRef, artifact_wandb.WandbArtifactRef
    ]
    branch_point: artifact_fs.BranchPointType


def get_merge_spec_uri(uri: str) -> typing.Optional[_MergeSpec]:
    try:
        return _get_merge_spec_unsafe(uri)
    except errors.WeaveMergeArtifactSpecError:
        return None


def _get_merge_spec_unsafe(uri: str) -> _MergeSpec:
    obj_uri = uris.WeaveURI.parse(uri)

    if not isinstance(
        obj_uri,
        (artifact_wandb.WeaveWBArtifactURI, artifact_local.WeaveLocalArtifactURI),
    ):
        raise errors.WeaveMergeArtifactSpecError(
            "Cannot merge artifact of type %s" % type(obj_uri)
        )

    head_ref = obj_uri.to_ref()
    obj_metadata = head_ref.artifact.read_metadata()
    if "branch_point" not in obj_metadata:
        raise errors.WeaveMergeArtifactSpecError(
            "Cannot merge into artifact without branch_point"
        )

    branch_point: artifact_fs.BranchPointType = obj_metadata["branch_point"]
    to_uri = uris.WeaveURI.parse(branch_point["original_uri"])
    if not isinstance(
        to_uri,
        (artifact_local.WeaveLocalArtifactURI, artifact_wandb.WeaveWBArtifactURI),
    ):
        raise errors.WeaveMergeArtifactSpecError(
            "Cannot merge into artifact of type %s" % type(obj_uri)
        )

    return _MergeSpec(
        to_uri=to_uri,
        head_ref=head_ref,
        branch_point=branch_point,
    )


def _merge(name) -> str:
    """Save the object, propagating changes back to the original artifact"""
    merge_spec = _get_merge_spec_unsafe(name)
    to_uri = merge_spec.to_uri
    head_ref = merge_spec.head_ref
    branch_point = merge_spec.branch_point

    to_ref = to_uri.to_ref()

    if isinstance(to_ref.artifact, artifact_wandb.WandbArtifact):
        original_artifact_hash = to_ref.artifact.commit_hash
    else:
        original_artifact_hash = to_ref.artifact.version
    if branch_point["commit"] != original_artifact_hash:
        raise errors.WeaveUnmergableArtifactsError(
            "Cannot merge artifacts: "
            "target artifact commit hash does not match branch point",
            f" original_artifact_hash: {original_artifact_hash}; ",
            f" branch_point: {branch_point['commit']}",
        )

    shared_branch_name = (
        None
        if to_uri != None
        and to_uri.version is not None
        and artifact_wandb.likely_commit_hash(to_uri.version)
        else to_uri.version
    )

    obj = head_ref.get()
    # Crucially, we must compute the weave type using type_of
    # instead of allowing _direct_publish and direct_save to compute
    # the weave type using type_of_with_refs. Merging is a special
    # operation and should not try to do ref tracking.
    weave_type = types.type_of(obj)

    ref: ref_base.Ref
    if isinstance(to_uri, artifact_wandb.WeaveWBArtifactURI):
        ref = storage._direct_publish(
            obj=obj,
            name=to_uri.name,
            assume_weave_type=weave_type,
            wb_project_name=to_uri.project_name,
            wb_entity_name=to_uri.entity_name,
            branch_name=shared_branch_name,
            _merge=True,
        )
    else:
        if to_uri.version is None:
            raise errors.WeaveInternalError(
                "Cannot merge into artifact without version"
            )

        ref = storage.direct_save(
            obj=head_ref.get(),
            name=to_uri.name,
            source_branch_name=shared_branch_name,
            branch_name=shared_branch_name,
            assume_weave_type=weave_type,
        )
    return ref.branch_uri


@op(
    pure=False,
    setter=_save,
    name="get",
    input_type={"uri": types.String()},
    output_type=types.Any(),
    refine_output_type=get_returntype,
)
def get(uri):
    ref = ref_base.Ref.from_str(uri)
    # if not ref.is_saved:
    #     return None
    try:
        res = ref.get()
    except errors.WeaveArtifactVersionNotFound:
        return None
    return res


def _save_with_default(
    get_uri,
    default,
    obj,
    root_args,
    make_new_type: typing.Callable[[types.Type], types.Type],
    mutation_record: object_context.MutationRecord,
):
    return _save(get_uri, obj, root_args, make_new_type, mutation_record)


@op(
    name="withdefault-getReturnType",
    input_type={"uri": types.String(), "default": types.Any()},
    output_type=types.TypeType(),
    hidden=True,
    pure=False,
)
def get_withdefault_returntype(uri, default):
    res = op_get_return_type(uri)
    if res != types.NoneType():
        return res
    return types.TypeRegistry.type_of(default)


@op(
    pure=False,
    setter=_save_with_default,
    name="withdefault-get",
    input_type={"uri": types.String(), "default": types.Any()},
    output_type=types.Any(),
    refine_output_type=get_withdefault_returntype,
    render_info={"type": "function"},
)
def get_withdefault(uri, default):
    ref = ref_base.Ref.from_str(uri)
    try:
        res = ref.get()
    except errors.WeaveArtifactVersionNotFound:
        return default
    return res


@op(
    pure=False,
    name="ref",
    render_info={"type": "function"},
    input_type={"uri": types.String()},
    # Only works on FilesystemArtifactRef right now, not generic.
    output_type=types.FilesystemArtifactRefType(),
)
def ref(uri):
    return ref_base.Ref.from_str(uri)


@op(
    pure=False,
    name="Ref-branch_point",
    input_type={"ref": types.FilesystemArtifactRefType()},
)
def ref_branch_point(ref) -> typing.Optional[artifact_fs.BranchPointType]:
    # TODO: execute automatically derefs. Need to not do that if input type is Ref!
    real_ref = storage._get_ref(ref)
    if not isinstance(real_ref, artifact_fs.FilesystemArtifactRef):
        raise errors.WeaveInternalError(
            "ref_branch_point only works on filesystem artifact refs"
        )
    return real_ref.branch_point


def execute_setter(node, value, action=None):
    if isinstance(node, graph.ConstNode):
        new_type = node.type
        if isinstance(new_type, types.Function):
            new_type = types.Function(new_type.input_types, value.type)
        return graph.ConstNode(new_type, value)
    else:
        # TODO: The OutputNode code path needs to do the mutation
        #     on node, instead of going back up whatever constructed
        #     Node
        raise errors.WeaveInternalError(
            "Execute setter path not implemented for ", node
        )


@op(
    # TODO: purity is a function of arguments in this special case.
    # E.g. if argument is "3 + 1", executing it is pure. If this matters
    # in practice, we can just hardcode in the engine.
    pure=False,
    setter=execute_setter,
    name="execute",
    input_type={"node": types.Function({}, types.Any())},
    output_type=lambda input_type: input_type["node"].output_type,
)
def execute(node):
    # Reenable compile, in case we're already compiling. This happens when a Node
    # is stored in an object (as in the way that the Panel architecture passes
    # information down to the server). Compile doesn't walk into objects. So we need
    # to compile here when we are about to use them.
    with compile.enable_compile():
        return weave_internal.use(node)


def mutate_op_body(
    self,
    root_args: typing.Any,
    make_new_value: typing.Callable[[typing.Any], typing.Any],
    make_new_type: typing.Callable[[types.Type], types.Type],
    mutation_record: object_context.MutationRecord,
):
    # This implements mutations. Note that its argument must be a
    # Node. You can call it like this:
    # weave.use(ops.set(weave_internal.const(csv[-1]["type"]), "YY"))

    self = compile.compile_fix_calls([self])[0]
    nodes = graph.linearize(self)
    if nodes is None:
        raise errors.WeaveInternalError("Set error")

    # Run through resolvers forward
    results = []
    op_inputs = []
    arg0 = (
        list(nodes[0].from_op.inputs.values())[0].val
        if len(nodes[0].from_op.inputs) > 0
        else None
    )
    for node in nodes:
        inputs = {}
        if len(node.from_op.inputs) > 0:
            arg0_name = list(node.from_op.inputs.keys())[0]
            inputs[arg0_name] = arg0
            for name, input_node in list(node.from_op.inputs.items())[1:]:
                if not isinstance(input_node, graph.ConstNode):
                    inputs[name] = storage.deref(weave_internal.use(input_node))
                    # TODO: I was raising here, but the way I'm handling
                    # default config in multi_distribution makes it necessary
                    # to handle this case. This solution is more general,
                    # but also more expensive. We should make use of the execution
                    # cache (mutations should probably be planned by the compiler)
                    # raise errors.WeaveInternalError("Set error")
                else:
                    inputs[name] = input_node.val
        op_inputs.append(inputs)
        op_def = registry_mem.memory_registry.get_op(node.from_op.name)
        arg0 = op_def.resolve_fn(**inputs)
        results.append(arg0)

    # Make the updates backwards
    res = make_new_value(arg0)

    for i, (node, inputs, result) in reversed(
        list(enumerate(zip(nodes, op_inputs, results)))
    ):
        op_def = registry_mem.memory_registry.get_op(node.from_op.name)
        if not op_def.setter:
            return res
            # TODO: we can't raise the error here. Some of the tests
            # rely on partial setter chains.
            # raise errors.WeaveInternalError(
            #     "Set error. No setter declared for op: %s" % node.from_op.name
            # )
        args = list(inputs.values())
        args.append(res)
        if i == 0 and (
            node.from_op.name == "get" or node.from_op.name == "withdefault-get"
        ):
            # TODO hardcoded get to take root_args. Should just check if available on setter.
            args.append(root_args)
            args.append(make_new_type)
            args.append(mutation_record)
        setter = op_def.setter
        try:
            setter = setter.func  # type: ignore
        except AttributeError:
            pass

        res = setter(*args)

    return res


@mutation
def set(
    self: graph.Node[typing.Any], val: typing.Any, root_args: typing.Any = None
) -> typing.Any:
    if root_args is None:
        root_args = {}

    if isinstance(self, graph.ConstNode):
        return val

    return mutate_op_body(
        self,
        root_args,
        lambda _: val,
        lambda _: types.TypeRegistry.type_of(val),
        object_context.MutationRecord("op-set", [self, val, root_args]),
    )


@mutation
def append(
    self: graph.Node[list], val: typing.Any, root_args: typing.Any = None
) -> typing.Any:
    if root_args is None:
        root_args = {}
    return mutate_op_body(
        self,
        root_args,
        lambda v: [val] if v == None else v + [val],
        lambda t: types.merge_types(t, types.List(types.TypeRegistry.type_of(val))),
        object_context.MutationRecord("op-append", [self, val, root_args]),
    )


@op(mutation=True, name="stream_table-log")
def stream_table_log(self: graph.Node, val: typing.Any) -> typing.Any:
    st_obj = weave_internal.use(self)
    from weave.monitoring import StreamTable

    if not isinstance(st_obj, StreamTable):
        raise errors.WeaveInternalError(
            "stream_table-log op must be called on a stream table"
        )

    # Tim: as part of weaveflow merge, i added underscores here. Not sure
    # how this ever worked before
    st = StreamTable(
        table_name=st_obj._table_name,
        project_name=st_obj._project_name,
        entity_name=st_obj._entity_name,
    )
    st.log(val)


def _get_uri_from_node(node: graph.Node[typing.Any], op_title: str) -> str:
    if not isinstance(node, graph.OutputNode):
        raise errors.WeaveInternalError(f"{op_title} target must be an OutputNode")

    if node.from_op.name != "get":
        raise errors.WeaveInternalError(f"{op_title} target must be a get")

    if not isinstance(node.from_op.inputs["uri"], graph.ConstNode):
        raise errors.WeaveInternalError(
            f"{op_title} op argument must be a const string"
        )

    return node.from_op.inputs["uri"].val


def _artifact_ref_from_uri(
    uri: str, op_title: str
) -> typing.Union[artifact_wandb.WandbArtifactRef, artifact_local.LocalArtifactRef]:
    obj_uri = uris.WeaveURI.parse(uri)
    if not isinstance(
        obj_uri,
        (artifact_wandb.WeaveWBArtifactURI, artifact_local.WeaveLocalArtifactURI),
    ):
        raise errors.WeaveInternalError(
            f"Cannot {op_title} artifact of type %s" % type(obj_uri)
        )

    return obj_uri.to_ref()


@mutation
def merge_artifact(
    self: graph.Node[typing.Any], root_args: typing.Any = None
) -> typing.Any:
    self = compile.compile_fix_calls([self])[0]
    return _merge(_get_uri_from_node(self, "Merge"))


@mutation
def delete_artifact(
    self: graph.Node[typing.Any], root_args: typing.Any = None
) -> typing.Any:
    head_ref = _artifact_ref_from_uri(_get_uri_from_node(self, "Delete"), "Delete")
    head_ref.artifact.delete()


@mutation
def undo_artifact(
    self: graph.Node[typing.Any], root_args: typing.Any = None
) -> typing.Any:
    head_ref = _artifact_ref_from_uri(_get_uri_from_node(self, "Undo"), "Undo")
    art = head_ref.artifact.undo()
    if art is None:
        raise errors.WeaveInternalError("Cannot undo artifact")
    base_uri = str(art.initial_uri_obj)
    if head_ref.path is not None:
        base_uri = base_uri + "/" + head_ref.path
    return base_uri


@mutation
def rename_artifact(
    self: graph.Node[typing.Any], new_name: str, root_args: typing.Any = None
) -> typing.Any:
    head_ref = _artifact_ref_from_uri(_get_uri_from_node(self, "Rename"), "Rename")
    head_ref.artifact.rename(new_name)


@op(
    mutation=True,
    hidden=True,
)
def publish_artifact(
    self: graph.Node[typing.Any],
    artifact_name: typing.Optional[str],
    project_name: typing.Optional[str],
    entity_name: typing.Optional[str],
    # root_args: typing.Any = None,
) -> str:
    head_ref = _artifact_ref_from_uri(_get_uri_from_node(self, "Publish"), "Publish")
    art_name = artifact_name or head_ref.artifact.name
    ref = storage._direct_publish(
        head_ref.get(),
        art_name,
        project_name,
        wb_entity_name=entity_name,
        metadata=head_ref.artifact.metadata.as_dict(),
    )
    return str(ref)


@weave_class(weave_type=types.Function)
class FunctionOps:
    @op(
        output_type=lambda input_type: input_type["self"].output_type,
    )
    def __call__(self, arg1: typing.Any):
        arg1_node = weave_internal.make_const_node(
            types.TypeRegistry.type_of(arg1), arg1
        )
        called = weave_internal.better_call_fn(self, arg1_node)  # type: ignore
        return weave_internal.use(called)


@weave_class(weave_type=types.RunType)
class Run:
    # TODO: the first arg type to these mutations should be Node[Run], not just
    # plain Run. They work for now but may not be callable from WeaveJS. We
    # should make weave_class automatically wrap the self arg in Node[] for
    # mutation ops!
    @mutation
    def set_state(self, state: str) -> None:
        r = typing.cast(runs.Run, self)
        set(r.state, state)

    @mutation
    def set_inputs(self, v: typing.Any) -> None:
        r = typing.cast(runs.Run, self)
        set(r.inputs, v)

    @mutation
    def set_output(self, v: typing.Any) -> None:
        r = typing.cast(runs.Run, self)
        set(r.output, v)

    @mutation
    def print(self, s: str) -> None:
        r = typing.cast(runs.Run, self)
        append(r.prints, s)

    @mutation
    def log(self, v: typing.Any) -> None:
        r = typing.cast(runs.Run, self)
        append(r.history, v)

    @op(
        input_type={"self": types.Function(output_type=types.RunType())},
    )
    def id(self) -> str:
        run = weave_internal.use(self)  # type: ignore

        # Hack: Since nullability is handled at execution time, it is entirely
        # possible for the run to be None, even though the type calls for
        # RunType. This is a safety check that any op accepting a function as
        # the first argument should account for.
        if run == None:
            return None  # type: ignore
        return run.id  # type: ignore

    @op(
        name="run-await",
        input_type={"self": types.Function(output_type=types.RunType())},
        output_type=lambda input_types: input_types["self"].output_type.output,
    )
    def await_final_output(self):
        # time.sleep(1000)
        sleep_mult = 1
        run_val = weave_internal.use(self)
        while run_val.state == "pending" or run_val.state == "running":
            sleep_time = 0.1 * sleep_mult
            if sleep_time > 3:
                sleep_time = 3
            sleep_mult *= 1.3
            time.sleep(sleep_time)
            run_val = weave_internal.use(self)

        return run_val.output



import weave

from .. import weave_types as types
from ..panels import panel_board
from .. import weave_internal
from .generator_templates import template_registry


panels = weave.panels

BOARD_INPUT_WEAVE_TYPE = types.List(
    types.TypedDict(
        {
            "timestamp": types.optional(types.Timestamp()),
            "_timestamp": types.Number(),
            "entity_name": types.optional(types.String()),
            "project_name": types.optional(types.String()),
            "queue_uri": types.optional(types.String()),
            "sweep_id": types.optional(types.String()),
            "run_id": types.optional(types.String()),
            "job": types.optional(types.String()),
            "trace_id": types.optional(types.String()),
            "state": types.optional(types.String()),
            "error": types.optional(types.String()),
            "metrics": types.TypedDict(
                {
                    "system": types.TypedDict(
                        {
                            "tpu": types.optional(types.Number()),
                            "cpu": types.optional(types.Number()),
                            "cpu_cores_util": types.optional(
                                types.List(types.Number())
                            ),
                            "gpu_cores_util": types.optional(
                                types.List(types.Number())
                            ),
                            "gpu_cores_mem": types.optional(types.List(types.Number())),
                            "memory": types.optional(types.Number()),
                            "disk": types.TypedDict(
                                {
                                    "in": types.optional(types.Number()),
                                    "out": types.optional(types.Number()),
                                }
                            ),
                            "proc": types.TypedDict(
                                {
                                    "memory": types.TypedDict(
                                        {
                                            "availableMB": types.optional(
                                                types.Number()
                                            ),
                                        }
                                    )
                                }
                            ),
                        }
                    )
                }
            ),
        }
    )
)


@weave.op(  # type: ignore
    name="py_board-observability",
    hidden=False,
    input_type={
        "input_node": types.Function(
            {},
            BOARD_INPUT_WEAVE_TYPE,
        )
    },
)
def observability(
    input_node,
) -> panels.Group:
    timestamp_col_name = "timestamp"
    num_buckets = 80

    now = weave.ops.datetime_now()

    dashboard = weave.panels.Group(
        layoutMode="grid",
        showExpressions=False,
        enableAddPanel=False,
        disableDeletePanel=True,
    )

    display_states = weave.ops.dict_(
        **{
            "running": "running",
            "failed": "failed",
            "crashed": "crashed",
            "finished": "finished",
            "starting": "pending",
            "failed_rqi": "failed rqi",
            "queued": "queued",
            "popped": "TO_REMOVE",  # gets filtered out
        }
    )

    varbar = panel_board.varbar(editable=False)
    source_data = input_node.map(
        lambda row: weave.ops.dict_(
            **{
                "timestamp": row[timestamp_col_name],
                "entity_name": row["entity_name"],
                "project_name": row["project_name"],
                "queue_uri": row["queue_uri"],
                "sweep_id": row["sweep_id"],
                "run_id": row["run_id"],
                "job": row["job"],
                "trace_id": row["trace_id"],
                "state": display_states.pick(row["state"]),
                "error": row["error"],
                "metrics": row["metrics"],
            }
        )
    )

    filter_fn = varbar.add(
        "filter_fn",
        weave_internal.define_fn(
            {"row": source_data.type.object_type},
            lambda row: weave_internal.const(True),
        ),
        hidden=True,
    )

    group_by_user = weave_internal.define_fn(
        {"row": source_data.type.object_type}, lambda row: row["entity_name"]
    )

    grouping_fn = varbar.add(
        "grouping_fn",
        weave_internal.define_fn(
            {"row": source_data.type.object_type}, lambda row: row["entity_name"]
        ),
        hidden=True,
    )

    seven_days_in_seconds = 60 * 60 * 24 * 7
    window_start = varbar.add(
        "window_start",
        weave.ops.from_number(now - seven_days_in_seconds),
        hidden=True,
    )
    window_end = varbar.add(
        "window_end",
        weave.ops.from_number(now),
        hidden=True,
    )

    filtered_range = varbar.add(
        "filtered_range",
        weave.ops.make_list(
            a=window_start,
            b=window_end,
        ),
        hidden=True,
    )

    user_zoom_range = varbar.add(
        "user_zoom_range",
        filtered_range,
        hidden=True,
    )
    varbar.add(
        "Time_range",
        panels.DateRange(user_zoom_range, domain=filtered_range),
    )
    bin_range = varbar.add(
        "bin_range", user_zoom_range.coalesce(filtered_range), hidden=True
    )

    window_data = source_data.filter(
        lambda row: weave.ops.Boolean.bool_and(
            row["state"] != "TO_REMOVE",  # remove hidden states
            weave.ops.Boolean.bool_and(
                row[timestamp_col_name] >= bin_range[0],
                row[timestamp_col_name] <= bin_range[1],
            ),
        )
    )
    varbar.add(
        "Grouping",
        panels.GroupingEditor(grouping_fn, node=window_data),
    )
    filtered_window_data = varbar.add(
        "filtered_window_data",
        window_data.filter(filter_fn),
        hidden=True,
    )

    varbar.add(
        "Filters",
        panels.FilterEditor(filter_fn, node=window_data),
    )

    colors_node = weave.ops.dict_(
        **{
            "running": "rgb(57, 126, 237)",
            "failed": "rgb(255, 122, 136)",
            "crashed": "rgb(255, 61, 90)",
            "finished": "rgb(0, 178, 110)",
            "pending": "rgb(125, 177, 250)",
            "failed rqi": "rgb(255, 199, 202)",
            "queued": "rgb(211, 215, 222)",
        }
    )

    state_color_func = weave_internal.define_fn(
        {"row": source_data.type.object_type},
        lambda row: colors_node.pick(row["state"]),
    )

    queued_time_data = varbar.add(
        "queued_time_data",
        filtered_window_data.filter(
            lambda row: weave.ops.Boolean.bool_or(
                row["state"] == "queued",
                row["state"] == "pending",
            ),
        ),
        hidden=True,
    )

    is_start_stop_state = weave_internal.define_fn(
        {"row": source_data.type.object_type},
        lambda row: weave.ops.Boolean.bool_or(
            row["state"] == "running",
            weave.ops.Boolean.bool_or(
                row["state"] == "finished",
                weave.ops.Boolean.bool_or(
                    row["state"] == "crashed",
                    row["state"] == "failed",
                ),
            ),
        ),
    )

    state_transitions_plot = panels.Plot(
        filtered_window_data,
        x=lambda row: row[timestamp_col_name].bin(
            weave.ops.timestamp_bins_nice(bin_range, num_buckets)
        ),
        x_title="Time",
        y=lambda row: row.count(),
        y_title="Count of transitions by state",
        label=lambda row: row["state"],
        tooltip=lambda row: weave.ops.dict_(
            **{
                "State": row["state"][0],
                "Count": row.count(),
                "Job (s)": weave.ops.join_to_str(row["job"].unique(), ","),
            }
        ),
        color_title="state",
        color=state_color_func.val,
        groupby_dims=["x", "label"],
        mark="bar",
        domain_x=user_zoom_range,
    )

    runs_grouped = weave.ops.List.groupby(queued_time_data, lambda row: row["trace_id"])
    runs_mapped = weave.ops.List.map(
        runs_grouped,
        lambda row: weave.ops.dict_(
            **{
                "trace_id": row["trace_id"][0],
                "entity_name": row["entity_name"][0],
                "project_name": row["project_name"][0],
                "run_id": row["run_id"][-1],
                "job": row["job"][-1],
                "enqueued": row[timestamp_col_name][0],
                "run_start": row[timestamp_col_name][-1],
                "duration": weave.ops.Number.__truediv__(
                    weave.ops.Number.__mul__(
                        weave.ops.timedelta_total_seconds(
                            weave.ops.datetime_sub(
                                row[timestamp_col_name].max(),
                                row[timestamp_col_name].min(),
                            ),
                        ),
                        1000,
                    ),
                    60,
                ),
            }
        ),
    )
    runs_mapped_filtered = varbar.add(
        "runs_mapped_filtered",
        weave.ops.List.filter(
            runs_mapped,
            lambda row: row["duration"] >= 0,
        ),
        hidden=True,
    )

    queued_time_plot = panels.Plot(
        runs_mapped_filtered,
        x=lambda row: row["run_start"].bin(
            weave.ops.timestamp_bins_nice(bin_range, num_buckets)
        ),
        x_title="Time",
        y=lambda row: row["duration"].sum(),
        y_title="Time spent queued (m)",
        label=lambda row: grouping_fn(row),
        tooltip=lambda row: weave.ops.dict_(
            **{
                "Job (s)": weave.ops.join_to_str(row["job"].unique(), ","),
                "User (s)": weave.ops.join_to_str(row["entity_name"].unique(), ","),
                "Project (s)": weave.ops.join_to_str(row["project_name"].unique(), ","),
                "Duration (m)": row["duration"].sum(),
                "Run count": row.count(),
            }
        ),
        color_title="Grouping",
        groupby_dims=["x", "label"],
        mark="bar",
        domain_x=user_zoom_range,
    )

    start_stop_states = varbar.add(
        "start_stop_states",
        weave.ops.List.filter(filtered_window_data, is_start_stop_state),
        hidden=True,
    )

    latest_runs_plot = panels.Plot(
        start_stop_states,
        x=lambda row: row[timestamp_col_name],
        x_title="Time",
        y_title="Job",
        y=lambda row: row["job"],
        tooltip=lambda row: weave.ops.dict_(
            **{
                "Job": row["job"][0],
                "User": row["entity_name"][0],
                "Team": row["queue_uri"][0].split(":").index(2),
                "Project": row["project_name"][0],
                "Run ID": row["run_id"][-1],
                "Status": row["state"][-1],
                "Duration (m)": weave.ops.Number.__truediv__(
                    weave.ops.Number.__mul__(
                        weave.ops.timedelta_total_seconds(
                            weave.ops.datetime_sub(
                                row[timestamp_col_name].max(),
                                row[timestamp_col_name].min(),
                            ),
                        ),
                        1000,
                    ),
                    60,
                ),
            }
        ),
        label=lambda row: row["run_id"],
        groupby_dims=["label"],
        color_title="User",
        color=group_by_user.val,
        mark="line",
        no_legend=True,
        domain_x=user_zoom_range,
    )

    jobs = varbar.add(
        "jobs",
        filtered_window_data.filter(is_start_stop_state),
        hidden=True,
    )
    jobs_table = panels.Table(jobs)  # type: ignore
    jobs_table.add_column(lambda row: row["run_id"], "Run", groupby=True)
    jobs_table.add_column(
        lambda row: weave.ops.run_ops.str_run_link(
            entity_name=row["queue_uri"][0].split(":").index(2),
            project_name=row["project_name"][0],
            name=row["run_id"][0],
        ),
        "Link",
        panel_def="link",
    )
    jobs_table.add_column(lambda row: row["job"][0], "Job")
    jobs_table.add_column(lambda row: row["entity_name"][0], "User")
    jobs_table.add_column(
        lambda row: row[timestamp_col_name][0], "Start Time", sort_dir="desc"
    )
    jobs_table.add_column(
        lambda row: row["state"][-1],
        "Current state",
    )
    jobs_table.add_column(
        lambda row: weave.ops.cond(
            weave.ops.dict_(
                a=row["state"][-1] != "running",
                b=row["state"][-1] == "running",
            ),
            weave.ops.dict_(
                a=weave.ops.Number.__truediv__(
                    weave.ops.timedelta_total_seconds(
                        weave.ops.datetime_sub(
                            row[timestamp_col_name].max(),
                            row[timestamp_col_name].min(),
                        )
                    ),
                    60,
                ),
                b=weave.ops.Number.__truediv__(
                    weave.ops.timedelta_total_seconds(
                        weave.ops.datetime_sub(
                            weave.ops.from_number(weave.ops.datetime_now()),
                            row[timestamp_col_name].min(),
                        )
                    ),
                    60,
                ),
            ),
        ),
        "Runtime (m)",
    )
    jobs_table.add_column(
        lambda row: weave.ops.cond(
            weave.ops.dict_(
                a=row["state"][-1] != "running",
                b=row["state"][-1] == "running",
            ),
            weave.ops.dict_(
                a=row["metrics"]["system"]["cpu_cores_util"][-1].avg(),
                b=weave.ops.make_const_node(types.NoneType(), None),
            ),
        ),
        "Avg. CPU %",
    )
    jobs_table.add_column(
        lambda row: weave.ops.cond(
            weave.ops.dict_(
                a=row["state"][-1] != "running",
                b=row["state"][-1] == "running",
            ),
            weave.ops.dict_(
                a=row["metrics"]["system"]["gpu_cores_util"][-1].avg(),
                b=weave.ops.make_const_node(types.NoneType(), None),
            ),
        ),
        "Avg. GPU %",
    )
    jobs_table.add_column(
        lambda row: weave.ops.cond(
            weave.ops.dict_(
                a=row["state"][-1] != "running",
                b=row["state"][-1] == "running",
            ),
            weave.ops.dict_(
                a=row["metrics"]["system"]["gpu_cores_mem"][-1].avg(),
                b=weave.ops.make_const_node(types.NoneType(), None),
            ),
        ),
        "Avg. GPU mem. %",
    )
    jobs_table.add_column(
        lambda row: weave.ops.cond(
            weave.ops.dict_(
                a=row["state"][-1] != "running",
                b=row["state"][-1] == "running",
            ),
            weave.ops.dict_(
                a=row["metrics"]["system"]["memory"][-1],
                b=weave.ops.make_const_node(types.NoneType(), None),
            ),
        ),
        "Avg. mem (MB)",
    )

    runs_table_data = weave.ops.List.groupby(
        filtered_window_data, lambda row: row["trace_id"]
    )
    runs_table_data_mapped = varbar.add(
        "runs_table_data",
        weave.ops.List.map(
            runs_table_data,
            lambda row: weave.ops.dict_(
                **{
                    "trace_id": row["trace_id"][0],
                    "user": row["entity_name"][0],
                    "run_ids": row["run_id"],
                    "states": row["state"],
                }
            ),
        ),
        hidden=True,
    )
    runs_table = panels.Table(runs_table_data_mapped)  # type: ignore
    runs_table.add_column(lambda row: row["user"], "User", groupby=True)
    runs_table.add_column(
        lambda row: row.count(),
        "Queued",
        sort_dir="desc",
    )
    runs_table.add_column(
        lambda row: row.filter(lambda row: row["states"][-1] == "failed rqi").count(),
        "Failed to start",
    )
    runs_table.add_column(
        lambda row: row.filter(lambda row: row["states"][2] == "running").count(),
        "Started",
    )
    runs_table.add_column(
        lambda row: row.filter(lambda row: row["states"][-1] == "running").count(),
        "Running",
    )
    runs_table.add_column(
        lambda row: row.filter(lambda row: row["states"][-1] == "finished").count(),
        "Finished",
    )
    runs_table.add_column(
        lambda row: row.filter(
            lambda row: weave.ops.Boolean.bool_or(
                row["states"][-1] == "crashed",
                row["states"][-1] == "failed",
            ),
        ).count(),
        "Crashed",
    )

    finished_runs = varbar.add(
        "finished_runs",
        filtered_window_data.filter(
            lambda row: weave.ops.Boolean.bool_or(
                row["state"] == "finished",
                weave.ops.Boolean.bool_or(
                    row["state"] == "crashed",
                    row["state"] == "failed",
                ),
            ),
        ),
        hidden=True,
    )

    runs_user_and_grouping_plot = panels.Plot(
        finished_runs,
        x=lambda row: row.count(),
        x_title="Count",
        y_title="Grouping",
        y=lambda row: grouping_fn(row),
        label=lambda row: grouping_fn(row),
        groupby_dims=["label"],
        mark="bar",
        no_legend=True,
    )

    metric_plot_data = weave.ops.List.groupby(
        start_stop_states, lambda row: row["trace_id"]
    )
    metric_plot_data_mapped = varbar.add(
        "metric_plot_data",
        weave.ops.List.map(
            metric_plot_data,
            lambda row: weave.ops.dict_(
                **{
                    "trace_id": row["trace_id"][0],
                    "run_id": row["run_id"][-1],
                    "entity_name": row["entity_name"][0],
                    "project_name": row["project_name"][0],
                    "job": row["job"][0],
                    "duration": weave.ops.Number.__truediv__(
                        weave.ops.Number.__mul__(
                            weave.ops.timedelta_total_seconds(
                                weave.ops.datetime_sub(
                                    row[timestamp_col_name].max(),
                                    row[timestamp_col_name].min(),
                                ),
                            ),
                            1000,
                        ),
                        60,
                    ),
                    "GPU util %": row["metrics"]["system"]["gpu_cores_util"][-1].avg(),
                    "CPU util %": row["metrics"]["system"]["cpu_cores_util"][-1].avg(),
                    "GPU memory %": row["metrics"]["system"]["gpu_cores_mem"][-1].avg(),
                    "Memory (MB)": row["metrics"]["system"]["memory"][-1],
                }
            ),
        ),
        hidden=True,
    )

    gpu_waste_by_user_plot = panels.Plot(
        metric_plot_data_mapped,
        x=lambda row: row["duration"],
        x_title="Run duration (minutes)",
        y_title="GPU utilization (%)",
        y=lambda row: row["GPU util %"],
        tooltip=lambda row: weave.ops.dict_(
            **{
                "Run ID": row["run_id"],
                "Project": row["project_name"],
                "User": row["entity_name"],
                "Job": row["job"],
                "Duration (m)": row["duration"],
                "GPU util %": row["GPU util %"],
            }
        ),
        color_title="Grouping",
        label=lambda row: grouping_fn(row),
        color=lambda row: grouping_fn(row),
        groupby_dims=[],
        mark="point",
        domain_y=weave_internal.make_const_node(types.List(types.Number()), [0, 100]),
    )

    cpu_waste_by_user_plot = panels.Plot(
        metric_plot_data_mapped,
        x=lambda row: row["duration"],
        x_title="Run duration (minutes)",
        y_title="CPU utilization (%)",
        y=lambda row: row["CPU util %"],
        tooltip=lambda row: weave.ops.dict_(
            **{
                "Run ID": row["run_id"],
                "Project": row["project_name"],
                "User": row["entity_name"],
                "Job": row["job"],
                "Duration (m)": row["duration"],
                "CPU util %": row["CPU util %"],
            }
        ),
        color_title="Grouping",
        label=lambda row: grouping_fn(row),
        color=lambda row: grouping_fn(row),
        groupby_dims=[],
        mark="point",
        domain_y=weave_internal.make_const_node(types.List(types.Number()), [0, 100]),
    )

    gpu_mem_by_user_plot = panels.Plot(
        metric_plot_data_mapped,
        x=lambda row: row["duration"],
        x_title="Run duration (minutes)",
        y_title="GPU memory (%)",
        y=lambda row: row["GPU memory %"],
        tooltip=lambda row: weave.ops.dict_(
            **{
                "Run ID": row["run_id"],
                "Project": row["project_name"],
                "User": row["entity_name"],
                "Job": row["job"],
                "Duration (m)": row["duration"],
                "GPU memory (%)": row["GPU memory %"],
            }
        ),
        color_title="Grouping",
        label=lambda row: grouping_fn(row),
        color=lambda row: grouping_fn(row),
        groupby_dims=[],
        mark="point",
        domain_y=weave_internal.make_const_node(types.List(types.Number()), [0, 100]),
    )

    memory_by_user_plot = panels.Plot(
        metric_plot_data_mapped,
        x=lambda row: row["duration"],
        x_title="Run duration (minutes)",
        y_title="Memory usage (MB)",
        y=lambda row: row["Memory (MB)"],
        tooltip=lambda row: weave.ops.dict_(
            **{
                "Run ID": row["run_id"],
                "Project": row["project_name"],
                "User": row["entity_name"],
                "Job": row["job"],
                "Duration (m)": row["duration"],
                "Memory (MB)": row["Memory (MB)"],
            }
        ),
        color_title="Grouping",
        label=lambda row: grouping_fn(row),
        color=lambda row: grouping_fn(row),
        groupby_dims=[],
        mark="point",
    )

    errors_table = panels.Table(  # type: ignore
        filtered_window_data.filter(
            weave_internal.define_fn(
                {"row": source_data.type.object_type},
                lambda row: row["error"] != None,
            )
        )
    )
    errors_table.add_column(
        lambda row: row[timestamp_col_name], "Timestamp", sort_dir="desc"
    )
    errors_table.add_column(lambda row: row["job"], "Job")
    errors_table.add_column(lambda row: row["error"], "Error", panel_def="object")

    # layout
    dashboard.add(
        "Job_run_status",
        state_transitions_plot,
        layout=panels.GroupPanelLayout(x=0, y=0, w=24, h=6),
    )
    dashboard.add(
        "Job_run_table",
        jobs_table,
        layout=panels.GroupPanelLayout(x=0, y=6, w=24, h=8),
    )
    dashboard.add(
        "Queued_time",
        queued_time_plot,
        layout=panels.GroupPanelLayout(x=0, y=14, w=24, h=6),
    )
    latest_runs_plot_selector = dashboard.add(
        "Job_runs",
        latest_runs_plot,
        layout=panels.GroupPanelLayout(x=0, y=20, w=24, h=6),
    )
    selected_jobs = panels.Table(latest_runs_plot_selector.selected_rows())  # type: ignore
    selected_jobs.add_column(
        lambda row: weave.ops.run_ops.str_run_link(
            entity_name=row["c_4.Team"],
            project_name=row["c_4.Project"],
            name=row["c_4.Run ID"],
        ),
        "Run",
        panel_def="link",
        groupby=True,
    )
    selected_jobs.add_column(lambda row: row["c_4.Job"][0], "Job")
    selected_jobs.add_column(lambda row: row["c_4.User"][0], "User")
    selected_jobs.add_column(lambda row: row["c_4.Project"][0], "Project")
    selected_jobs.add_column(lambda row: row["c_4.Status"][-1], "Status")
    selected_jobs.add_column(lambda row: row["c_2"].min(), "Start")
    selected_jobs.add_column(lambda row: row["c_2"].max(), "Stop")
    selected_jobs.add_column(lambda row: row["c_4.Duration (m)"][0], "Duration (m)")

    dashboard.add(
        "Selected_job_runs",
        selected_jobs,
        layout=panels.GroupPanelLayout(x=0, y=26, w=24, h=10),
    )
    dashboard.add(
        "Runs_by_user",
        runs_table,
        layout=panels.GroupPanelLayout(x=10, y=36, w=14, h=8),
    )
    dashboard.add(
        "GPU_use_by_job_run",
        gpu_waste_by_user_plot,
        layout=panels.GroupPanelLayout(x=0, y=44, w=12, h=8),
    )
    dashboard.add(
        "CPU_use_by_job_run",
        cpu_waste_by_user_plot,
        layout=panels.GroupPanelLayout(x=12, y=44, w=12, h=8),
    )
    dashboard.add(
        "GPU_memory_by_job_run",
        gpu_mem_by_user_plot,
        layout=panels.GroupPanelLayout(x=0, y=52, w=12, h=8),
    )
    dashboard.add(
        "System_memory_by_job_run",
        memory_by_user_plot,
        layout=panels.GroupPanelLayout(x=12, y=52, w=12, h=8),
    )
    dashboard.add(
        "Finished_runs_by_grouping",
        runs_user_and_grouping_plot,
        layout=panels.GroupPanelLayout(x=0, y=34, w=10, h=8),
    )
    dashboard.add(
        "Errors",
        errors_table,
        layout=panels.GroupPanelLayout(x=0, y=60, w=24, h=8),
    )

    return panels.Board(vars=varbar, panels=dashboard, editable=False)


template_registry.register(
    "py_board-observability",
    "Launch Observability",
    "Seed a board with an observability view to track Launch queue performance.",
)



import click
import os
import time
import typing
import sys
from weave import __version__

# from .model_server import app
from .urls import BROWSE3_PATH
from . import api
from . import uris
from . import artifact_wandb
from .deploy import gcp as google

# TODO: does this work?
os.environ["PYTHONUNBUFFERED"] = "1"


@click.group()
@click.version_option(version=__version__)
def cli() -> None:
    pass


@cli.command("ui", help="Start the weave UI.")
def start_ui() -> None:
    print("Starting server...")
    try:
        from weave import server
    except ModuleNotFoundError:
        print("Run 'pip install weave[engine]' to use the local server.")
        sys.exit(1)
    serv = server.HttpServer(port=3000)  # type: ignore
    serv.start()
    print("Server started")
    print(f"http://localhost:3000/{BROWSE3_PATH}")
    while True:
        time.sleep(10)


@cli.command(help="Serve weave models.")
@click.argument("model_ref")
@click.option("--method", help="Method name to serve.")
@click.option("--project", help="W&B project name.")
@click.option("--env", default="development", help="Environment to tag the model with.")
@click.option(
    "--auth-entity", help="Enforce authorization for this entity using W&B API keys."
)
@click.option("--port", default=9996, type=int)
def serve(
    model_ref: str,
    method: typing.Optional[str],
    auth_entity: typing.Optional[str],
    project: str,
    env: str,
    port: int,
) -> None:
    parsed_uri = uris.WeaveURI.parse(model_ref)
    if not isinstance(parsed_uri, artifact_wandb.WeaveWBArtifactURI):
        raise ValueError(f"Expected a weave artifact uri, got {parsed_uri}")
    parsed_ref = parsed_uri.to_ref()
    ref_project = parsed_uri.project_name
    project_override = project or os.getenv("PROJECT_NAME")
    if project_override:
        print(f"Logging to project different from {ref_project}")
        project = project_override
    else:
        project = ref_project

    api.init(project)
    # TODO: provide more control over attributes
    with api.attributes({"env": env}):
        api.serve(
            parsed_ref, method_name=method or None, auth_entity=auth_entity, port=port
        )


@cli.group(help="Deploy weave models.")
def deploy() -> None:
    pass


@deploy.command(help="Deploy to GCP.")
@click.argument("model_ref")
@click.option("--method", help="Method name to serve.")
@click.option("--project", help="W&B project name.")
@click.option("--gcp-project", help="GCP project name.")
@click.option(
    "--auth-entity", help="Enforce authorization for this entity using W&B API keys."
)
@click.option("--service-account", help="GCP service account.")
@click.option("--dev", is_flag=True, help="Run the function locally.")
def gcp(
    model_ref: str,
    method: str,
    project: str,
    gcp_project: str,
    auth_entity: str,
    service_account: str,
    dev: bool = False,
) -> None:
    if dev:
        print(f"Developing model {model_ref}...")
        google.develop(model_ref, model_method=method, auth_entity=auth_entity)
        return
    print(f"Deploying model {model_ref}...")
    if auth_entity is None:
        print(
            "WARNING: No --auth-entity specified.  This endpoint will be publicly accessible."
        )
    try:
        google.deploy(
            model_ref,
            model_method=method,
            wandb_project=project,
            auth_entity=auth_entity,
            gcp_project=gcp_project,
            service_account=service_account,
        )
    except ValueError as e:
        if os.getenv("DEBUG") == "true":
            raise e
        else:
            raise click.ClickException(
                str(e) + "\nRun with DEBUG=true to see full exception."
            )
    print("Model deployed")


@deploy.command(help="Deploy to Modal Labs.")
@click.argument("model_ref")
@click.option("--project", help="W&B project name.")
@click.option(
    "--auth-entity", help="Enforce authorization for this entity using W&B API keys."
)
@click.option("--dev", is_flag=True, help="Run the function locally.")
def modal(model_ref: str, project: str, auth_entity: str, dev: bool = False) -> None:
    from .deploy import modal as mdp

    if dev:
        print(f"Developing model {model_ref}...")
        mdp.develop(model_ref, auth_entity=auth_entity)
        return
    print(f"Deploying model {model_ref}...")
    if auth_entity is None:
        print(
            "WARNING: No --auth-entity specified.  This endpoint will be publicly accessible."
        )
    try:
        mdp.deploy(model_ref, wandb_project=project, auth_entity=auth_entity)
    except ValueError as e:
        if os.getenv("DEBUG") == "true":
            raise e
        else:
            raise click.ClickException(
                str(e) + "\nRun with DEBUG=true to see full exception."
            )
    print("Model deployed")


if __name__ == "__main__":
    cli()



import dataclasses
import typing
import wandb
from wandb import Artifact

from weave import wandb_client_api
from weave.wandb_interface.wandb_lite_run import InMemoryLazyLiteRun
from .. import engine_trace
from wandb.apis.public import api as wb_public


def artifact_commithash_by_digest(
    entity_name: str, project_name: str, artifact_name: str, digest: str
) -> typing.Optional[str]:
    query = wb_public.gql(
        """
    query ArtifactVersionFromDigest(
        $entityName: String!,
        $projectName: String!,
        $artName: String!,
        $digest: String!,
    ) {
        project(name: $projectName, entityName: $entityName) {
            artifactCollection(name: $artName) {
                artifactMembership(aliasName: $digest) {
                    commitHash
                    artifact {
                        id
                    }
                }
            }
        }
    }
    """
    )
    res = wandb_client_api.wandb_public_api().client.execute(
        query,
        # variable_values={"id": hex_to_b64_id(server_id)},
        variable_values={
            "entityName": entity_name,
            "projectName": project_name,
            "artName": artifact_name,
            "digest": digest,
        },
    )
    project = res.get("project")
    if project is None:
        return None
    artifact_collection = project.get("artifactCollection")
    if artifact_collection is None:
        return None
    artifact_membership = artifact_collection.get("artifactMembership")
    if artifact_membership is None:
        return None
    return artifact_membership.get("commitHash")


@dataclasses.dataclass
class WeaveWBArtifactURIComponents:
    entity_name: str
    project_name: str
    artifact_name: str
    version_str: str


def write_artifact_to_wandb(
    artifact: wandb.Artifact,
    project_name: str,
    entity_name: typing.Optional[str] = None,
    additional_aliases: list = [],
    *,
    _lite_run: typing.Optional[InMemoryLazyLiteRun] = None,
    artifact_collection_exists: bool = False,
) -> WeaveWBArtifactURIComponents:
    artifact_name = artifact.name

    tracer = engine_trace.tracer()  # type: ignore
    # Here we get the default entity if none is provided.
    # When we support saving dashboards to target entities,
    # we may want to rework this code path so that the caller
    # must provide an entity name.
    entity_name = entity_name or wandb_client_api.wandb_public_api().default_entity

    if entity_name is None:
        raise ValueError("entity_name must be provided")

    # Finalize the artifact, and return early if the digest already exists on the
    # server.
    artifact.finalize()
    existing_commit_hash = artifact_commithash_by_digest(
        entity_name=entity_name,
        project_name=project_name,
        artifact_name=artifact_name,
        digest=artifact.digest,
    )
    if existing_commit_hash is not None:
        return WeaveWBArtifactURIComponents(
            entity_name=entity_name,
            project_name=project_name,
            artifact_name=artifact_name,
            version_str=existing_commit_hash,
        )

    if _lite_run is None:
        lite_run = InMemoryLazyLiteRun(
            entity_name,
            project_name,
            group="weave_artifact_pushers",
            _hide_in_wb=True,
        )
    else:
        lite_run = _lite_run

    with tracer.trace("Logging artifact"):
        res = lite_run.log_artifact(
            artifact,
            additional_aliases,
            artifact_collection_exists,
        )

    with tracer.trace("Finish lite_run"):
        lite_run.finish()

    if res is not None:
        art = Artifact._from_id(res["id"], wandb_client_api.wandb_public_api().client)
        if art is not None:
            commit_hash = art.commit_hash

    # Return the URI of the artifact
    return WeaveWBArtifactURIComponents(
        entity_name=entity_name,
        project_name=project_name,
        artifact_name=artifact.name,
        version_str=commit_hash,
    )



import weave
from ..panels import panel_board


# This is not yet general, it describes a board for a specific
# formulation of a text extraction problem


def eval_board(dataset, eval_result0, eval_result1):  # type: ignore
    varbar = panel_board.varbar()
    dataset_var = varbar.add("dataset", dataset)
    baseline_eval_result = varbar.add("baseline_eval_result", eval_result0)
    candidate_eval_result = varbar.add("candidate_eval_result", eval_result1)

    summary = varbar.add(
        "summary",
        weave.ops.make_list(
            a=weave.ops.TypedDict.merge(
                weave.ops.dict_(name="baseline"), baseline_eval_result["summary"]
            ),
            b=weave.ops.TypedDict.merge(
                weave.ops.dict_(name="candidate"), candidate_eval_result["summary"]
            ),
        ),
    )

    weave.ops.make_list(
        a=baseline_eval_result["eval_table"], b=baseline_eval_result["eval_table"]
    )

    concatted_evals = varbar.add(
        "concatted_evals",
        weave.ops.List.concat(
            weave.ops.make_list(
                a=baseline_eval_result["eval_table"].map(
                    lambda row: weave.ops.TypedDict.merge(
                        weave.ops.dict_(name="baseline"), row
                    )
                ),
                b=candidate_eval_result["eval_table"].map(
                    lambda row: weave.ops.TypedDict.merge(
                        weave.ops.dict_(name="candidate"), row
                    )
                ),
            )
        ),
    )

    # join evals together first
    joined_evals = varbar.add(
        "joined_evals",
        weave.ops.join_all(
            weave.ops.make_list(
                a=baseline_eval_result["eval_table"],
                b=candidate_eval_result["eval_table"],
            ),
            lambda row: row["dataset_id"],
            False,
        ),
    )

    # then join dataset to evals
    dataset_evals = varbar.add(
        "dataset_evals",
        weave.ops.join_2(
            dataset_var,
            joined_evals,
            lambda row: row["id"],
            lambda row: row["dataset_id"][0],
            "dataset",
            "evals",
            False,
            False,
        ),
    )

    main = weave.panels.Group(
        layoutMode="grid",
        showExpressions=True,
        enableAddPanel=True,
    )

    #### Run/config info TODO

    #### Summary info

    main.add(
        "avg_f1",
        weave.panels.Plot(
            summary,
            x=lambda row: row["avg_f1"],
            y=lambda row: row["name"],
            color=lambda row: row["name"],
        ),
        layout=weave.panels.GroupPanelLayout(x=0, y=0, w=12, h=4),
    )

    main.add(
        "latency",
        weave.panels.Plot(
            concatted_evals,
            x=lambda row: row["summary"]["latency"],
            y=lambda row: row["name"],
            color=lambda row: row["name"],
            mark="boxplot",
        ),
        layout=weave.panels.GroupPanelLayout(x=12, y=0, w=12, h=4),
    )

    main.add(
        "field_name_f1",
        weave.panels.Plot(
            summary,
            x=lambda row: row["field_name.f1"],
            y=lambda row: row["name"],
            color=lambda row: row["name"],
        ),
        layout=weave.panels.GroupPanelLayout(x=0, y=4, w=8, h=4),
    )
    main.add(
        "field_shares_f1",
        weave.panels.Plot(
            summary,
            x=lambda row: row["field_shares.f1"],
            y=lambda row: row["name"],
            color=lambda row: row["name"],
        ),
        layout=weave.panels.GroupPanelLayout(x=8, y=4, w=8, h=4),
    )
    main.add(
        "field_directors_f1",
        weave.panels.Plot(
            summary,
            x=lambda row: row["field_directors.f1"],
            y=lambda row: row["name"],
            color=lambda row: row["name"],
        ),
        layout=weave.panels.GroupPanelLayout(x=16, y=4, w=8, h=4),
    )

    # ct = main.add('concat_t', concatted_evals, layout=weave.panels.GroupPanelLayout(x=0, y=4, w=24, h=12))
    # main.add('dataset_table', dataset)
    # main.add('joined_evals', joined_evals)
    # main.add(
    #     "dataset_evals",
    #     dataset_evals,
    #     layout=weave.panels.GroupPanelLayout(x=0, y=4, w=24, h=6),
    # )

    ##### Example details

    # more ideas: show examples that all got wrong, or that are confusing

    # facet_f1 = weave.panels.Facet(
    #     dataset_evals,
    #     x=lambda row: row["evals.summary"][0]["f1"],
    #     y=lambda row: row["evals.summary"][1]["f1"],
    #     select=lambda row: row.count(),
    # )

    # f1_comparison = main.add(
    #     "f1_comparison",
    #     facet_f1,
    #     layout=weave.panels.GroupPanelLayout(x=0, y=8, w=12, h=6),
    # )

    facet_correct = weave.panels.Facet(
        dataset_evals,
        x=lambda row: row["evals.summary"][0]["correct"],
        x_title="baseline correct",
        y=lambda row: row["evals.summary"][1]["correct"],
        y_title="candidate correct",
        select=lambda row: row.count(),
    )

    correct_comparison = main.add(
        "correct_comparison",
        facet_correct,
        layout=weave.panels.GroupPanelLayout(x=0, y=8, w=12, h=6),
    )

    main.add(
        "help",
        weave.panels.PanelString(
            "Click a cell in in the panel to the left to load examples for that cell.\n\nClick a row number in the table below to see details for that row.",
            mode="markdown",
        ),
        layout=weave.panels.GroupPanelLayout(x=12, y=8, w=12, h=6),
    )
    # main.add(
    #     "example_latencies",
    #     weave.panels.Plot(
    #         dataset_evals,
    #         x=lambda row: row["evals.summary"]["latency"][0],
    #         y=lambda row: row["evals.summary"]["latency"][1],
    #     ),
    #     layout=weave.panels.GroupPanelLayout(x=12, y=8, w=12, h=6),
    # )

    sel_ex_table = weave.panels.Table(correct_comparison.selected())
    sel_ex_table.config.rowSize = 2
    sel_ex_table.add_column(lambda row: row["dataset.id"], "id")
    sel_ex_table.add_column(lambda row: row["dataset.example"], "example")
    sel_ex_table.add_column(lambda row: row["dataset.label.name"], "label.name")
    sel_ex_table.add_column(
        lambda row: weave.ops.dict_(
            baseline=row["evals.result"][0]["name"],
            candidate=row["evals.result"][1]["name"],
        ),
        "result.name",
    )
    sel_ex_table.add_column(lambda row: row["dataset.label.shares"], "label.shares")
    sel_ex_table.add_column(
        lambda row: weave.ops.dict_(
            baseilne=row["evals.result"][0]["shares"].toString(),
            candidate=row["evals.result"][1]["shares"].toString(),
        ),
        "result.shares",
    )
    sel_ex_table.add_column(
        lambda row: row["dataset.label.directors"], "label.directors"
    )
    sel_ex_table.add_column(
        lambda row: weave.ops.dict_(
            baseilne=row["evals.result"][0]["directors"].toString(),
            candidate=row["evals.result"][1]["directors"].toString(),
        ),
        "result.directors",
    )
    sel_ex_table.add_column(
        lambda row: weave.ops.dict_(
            baseilne=row["evals.summary"][0]["latency"],
            candidate=row["evals.summary"][1]["latency"],
        ),
        "latency",
    )

    selected_examples = main.add(
        "selected_examples",
        sel_ex_table,
        layout=weave.panels.GroupPanelLayout(x=0, y=14, w=24, h=12),
    )

    main.add(
        "baseilne_detail",
        selected_examples.active_data()["evals.summary"][0],
        layout=weave.panels.GroupPanelLayout(x=0, y=26, w=12, h=8),
    )

    main.add(
        "candidate_detail",
        selected_examples.active_data()["evals.summary"][1],
        layout=weave.panels.GroupPanelLayout(x=12, y=26, w=12, h=8),
    )

    return weave.panels.Board(vars=varbar, panels=main)



import typing
import json
from ..gql_op_plugin import wb_gql_op_plugin
from ..api import op
from . import wb_domain_types as wdt
from ..language_features.tagging.make_tag_getter_op import make_tag_getter_op
from .wandb_domain_gql import (
    gql_prop_op,
    gql_direct_edge_op,
    gql_connection_op,
    gql_root_op,
    make_root_op_gql_op_output_type,
)

from .. import weave_types as types
from .. import errors
from .. import input_provider

# Section 1/6: Tag Getters
get_project_tag = make_tag_getter_op("project", wdt.ProjectType, op_name="tag-project")

# Section 2/6: Root Ops
# The output in this case is stored so it can be exported in __init__.py (can probably remove this
# as I believe it is only used in testing)
project = gql_root_op(
    "root-project",
    "project",
    wdt.ProjectType,
    {
        "entityName": types.String(),
        "projectName": types.String(),
    },
    lambda inputs: f'name: {inputs["projectName"]}, entityName: {inputs["entityName"]}',
)


@op(
    name="root-allProjectsGQLResolver",
    input_type={"gql_result": types.TypedDict({})},
    output_type=types.List(wdt.ProjectType),
    hidden=True,
)
def root_all_projects_gql_resolver(gql_result):
    return [
        wdt.Project.from_keys(project["node"])
        for project in gql_result["instance"]["projects_500"]["edges"]
    ]


@op(
    name="root-allProjects",
    input_type={},
    output_type=types.List(wdt.ProjectType),
    plugins=wb_gql_op_plugin(
        lambda inputs, inner: f"""
    instance {{
        projects_500: projects(limit: 500) {{
            edges {{
                node {{
                    {wdt.Project.REQUIRED_FRAGMENT}
                    {inner}
                }}
            }}
        }}
    }}
    """,
        is_root=True,
        root_resolver=root_all_projects_gql_resolver,
        gql_op_output_type=make_root_op_gql_op_output_type(
            "projects_500", lambda inputs: "", wdt.ProjectType
        ),
    ),
)
def root_all_projects():
    raise errors.WeaveGQLCompileError(
        "root-allProjects should not be executed directly. If you see this error, it is a bug in the Weave compiler."
    )


# Section 3/6: Attribute Getters
name = gql_prop_op("project-name", wdt.ProjectType, "name", types.String())
gql_prop_op(
    "project-createdAt",
    wdt.ProjectType,
    "createdAt",
    types.Timestamp(),
)
gql_prop_op(
    "project-updatedAt",
    wdt.ProjectType,
    "updatedAt",
    types.Timestamp(),
)

gql_prop_op(
    "project-internalId",
    wdt.ProjectType,
    "id",
    types.String(),
)

# Section 4/6: Direct Relationship Ops
gql_direct_edge_op(
    "project-run",
    wdt.ProjectType,
    "run",
    wdt.RunType,
    {
        "runName": types.String(),
    },
    lambda inputs: f'name: {inputs["runName"]}',
)

gql_direct_edge_op(
    "project-entity",
    wdt.ProjectType,
    "entity",
    wdt.EntityType,
)

gql_direct_edge_op(
    "project-runQueues",
    wdt.ProjectType,
    "runQueues",
    wdt.RunQueueType,
    is_many=True,
)

gql_direct_edge_op(
    "project-artifactType",
    wdt.ProjectType,
    "artifactType",
    wdt.ArtifactTypeType,
    {
        "artifactType": types.String(),
    },
    lambda inputs: f'name: {inputs["artifactType"]}',
)

gql_direct_edge_op(
    "project-artifactVersion",
    wdt.ProjectType,
    "artifact",
    wdt.ArtifactVersionType,
    {
        "artifactName": types.String(),
        "artifactVersionAlias": types.String(),
    },
    lambda inputs: f'name: {json.dumps(inputs.raw["artifactName"] + ":" + inputs.raw["artifactVersionAlias"])}',
)

gql_direct_edge_op(
    "project-artifact",
    wdt.ProjectType,
    "artifactCollection",
    wdt.ArtifactCollectionType,
    {
        "artifactName": types.String(),
    },
    lambda inputs: f'name: {inputs["artifactName"]}',
)

# Section 5/6: Connection Ops
gql_connection_op(
    "project-artifactTypes",
    wdt.ProjectType,
    "artifactTypes",
    wdt.ArtifactTypeType,
    {},
    # lambda inputs: "first: 10",
)

gql_connection_op(
    "project-runs",
    wdt.ProjectType,
    "runs",
    wdt.RunType,
    {},
    lambda inputs: "first: 100",
)


gql_connection_op(
    "project-filteredRuns",
    wdt.ProjectType,
    "runs",
    wdt.RunType,
    {
        "filter": types.String(),
        "order": types.String(),
    },
    lambda inputs: f'first: 100, filters: {inputs["filter"]}, order: {inputs["order"]}',
)


# Section 6/6: Non Standard Business Logic Ops
# Note: with project-link, the gql variables are already part fo the required fragment
# so there is nothing more to do here.
@op(
    name="project-link",
    plugins=wb_gql_op_plugin(
        lambda inputs, inner: """
            entity {
                id
                name
            }
        """
    ),
)
def link(project: wdt.Project) -> wdt.Link:
    return wdt.Link(project["name"], f"{project['entity']['name']}/{project['name']}")


def _project_artifacts_gql_op_output_type(
    inputs: input_provider.InputProvider, input_type: types.Type
) -> types.Type:
    return types.List(
        wdt.ArtifactCollectionType.with_keys(
            typing.cast(typing.Any, input_type)
            .keys["artifactTypes_100"]["edges"]
            .object_type["node"]["artifactCollections_100"]["edges"]
            .object_type["node"]
            .property_types
        )
    )


@op(
    name="project-artifacts",
    output_type=lambda input_types: types.List(wdt.ArtifactCollectionType),
    plugins=wb_gql_op_plugin(
        lambda inputs, inner: f"""
            artifactTypes_100: artifactTypes(first: 100) {{
                edges {{
                    node {{
                        id
                        artifactCollections_100: artifactCollections(first: 100) {{
                            edges {{
                                node {{
                                    {wdt.ArtifactCollection.REQUIRED_FRAGMENT}
                                    {inner}
                                }}
                            }}
                        }}
                    }}
                }}
            }}
        """,
        gql_op_output_type=_project_artifacts_gql_op_output_type,
    ),
)
def artifacts(
    project: wdt.Project,
):
    return [
        wdt.ArtifactCollection.from_keys(edge["node"])
        for typeEdge in project["artifactTypes_100"]["edges"]
        for edge in typeEdge["node"]["artifactCollections_100"]["edges"]
    ]



import typing

from ..gql_op_plugin import wb_gql_op_plugin
from ..api import op
from . import wb_domain_types as wdt
from .. import weave_types as types

from .wandb_domain_gql import (
    gql_prop_op,
    gql_direct_edge_op,
    gql_connection_op,
    gql_root_op,
    _make_alias,
)

# Section 1/6: Tag Getters
# None

# Section 2/6: Root Ops
# None


# Section 3/6: Attribute Getters
# op_artifact_type_name is written in the plain style
# because the attribute is part of the required fragment
@op(name="artifactType-name")
def op_artifact_type_name(artifactType: wdt.ArtifactType) -> str:
    return artifactType["name"]


# Section 4/6: Direct Relationship Ops
# None

# Section 5/6: Connection Ops
gql_connection_op(
    "artifactType-artifacts",
    wdt.ArtifactTypeType,
    "artifactCollections",
    wdt.ArtifactCollectionType,
    {},
    lambda inputs: "first: 100",
)

# Section 6/6: Non Standard Business Logic Ops
# This is a horrible op due to the double limits - we should remove this.
first_100_collections_alias = _make_alias("first: 100", prefix="artifactCollections")
first_100_artifacts_alias = _make_alias("first: 100", prefix="artifacts")


@op(
    name="artifactType-artifactVersions",
    plugins=wb_gql_op_plugin(
        lambda inputs, inner: f"""
    {first_100_collections_alias}: artifactCollections(first: 100) {{
        edges {{
            node {{
                {first_100_artifacts_alias}: artifacts(first: 100) {{
                    edges {{
                        node {{
                            {wdt.ArtifactVersion.REQUIRED_FRAGMENT}
                            {inner}
                        }}
                    }}
                }}
            }}
        }}
    }}""",
        gql_op_output_type=lambda inputs, input_type: types.List(
            wdt.ArtifactVersionType.with_keys(
                typing.cast(typing.Any, input_type)
                .keys[first_100_collections_alias]["edges"]
                .object_type["node"][first_100_artifacts_alias]["edges"]
                .object_type["node"]
                .property_types
            )
        ),
    ),
)
def artifact_versions(
    artifactType: wdt.ArtifactType,
) -> list[wdt.ArtifactVersion]:
    res = []
    for artifactCollectionEdge in artifactType[first_100_collections_alias]["edges"]:
        for artifactEdge in artifactCollectionEdge["node"][first_100_artifacts_alias][
            "edges"
        ]:
            res.append(wdt.ArtifactVersion.from_keys(artifactEdge["node"]))
    return res



import typing
import graphql

from . import environment
from . import wandb_client_api

_GQL_SCHEMA_CACHE: dict[typing.Optional[str], graphql.GraphQLSchema] = {}


def gql_schema() -> graphql.GraphQLSchema:
    schema_path = environment.gql_schema_path()
    gql_schema = _GQL_SCHEMA_CACHE.get(schema_path)

    if gql_schema is None:
        if schema_path is not None:
            with open(schema_path, "r") as f:
                schema_str = f.read()

            gql_schema = graphql.build_schema(schema_str)
        else:
            gql_schema = wandb_client_api.introspect_server_schema()

        _GQL_SCHEMA_CACHE[schema_path] = gql_schema

    return gql_schema



"""These are the top-level functions in the `import weave` namespace.
"""

import time
import typing
import os
import contextlib
import dataclasses
from typing import Any

from . import urls

from . import graph as _graph
from .graph import Node

# If this is not imported, serialization of Weave Nodes is incorrect!
from . import graph_mapper as _graph_mapper

from . import storage as _storage
from . import ref_base as _ref_base
from . import artifact_wandb as _artifact_wandb
from . import wandb_api as _wandb_api

from . import weave_internal as _weave_internal
from . import errors as _errors

from . import util as _util

from . import context as _context
from . import context_state as _context_state
from . import run as _run
from . import weave_init as _weave_init
from . import graph_client as _graph_client
from . import graph_client_local as _graph_client_local
from . import graph_client_wandb_art_st as _graph_client_wandb_art_st
from . import graph_client_context as _graph_client_context
from weave.monitoring import monitor as _monitor

# exposed as part of api
from . import weave_types as types

# needed to enable automatic numpy serialization
from . import types_numpy as _types_numpy

from . import errors
from .decorators import weave_class, op, mutation, type

from .op_def import OpDef
from . import usage_analytics
from .context import (
    use_fixed_server_port,
    use_frontend_devmode,
    # eager_execution,
    use_lazy_execution,
)

from .panel import Panel

from .arrow.list_ import ArrowWeaveList as WeaveList


def save(node_or_obj, name=None):
    from .ops_primitives.weave_api import save, get

    if isinstance(node_or_obj, _graph.Node):
        return save(node_or_obj, name=name)
    else:
        # If the user does not provide a branch, then we explicitly set it to
        # the default branch, "latest".
        branch = None
        name_contains_branch = name is not None and ":" in name
        if not name_contains_branch:
            branch = "latest"
        ref = _storage.save(node_or_obj, name=name, branch=branch)
        if name is None:
            # if the user didn't provide a name, the returned reference
            # will be to the specific version
            uri = ref.uri
        else:
            # otherwise the reference will be to whatever branch was provided
            # or the "latest" branch if only a name was provided.
            uri = ref.branch_uri
        return get(str(uri))


def get(ref_str):
    obj = _storage.get(ref_str)
    ref = typing.cast(_ref_base.Ref, _storage._get_ref(obj))
    return _weave_internal.make_const_node(ref.type, obj)


def use(nodes, client=None):
    usage_analytics.use_called()
    if client is None:
        client = _context.get_client()
    return _weave_internal.use(nodes, client)


def _get_ref(obj):
    if isinstance(obj, _storage.Ref):
        return obj
    ref = _storage.get_ref(obj)
    if ref is None:
        raise _errors.WeaveApiError("obj is not a weave object: %s" % obj)
    return ref


def versions(obj):
    if isinstance(obj, _graph.ConstNode):
        obj = obj.val
    elif isinstance(obj, _graph.OutputNode):
        obj = use(obj)
    ref = _get_ref(obj)
    return ref.versions()  # type: ignore


def expr(obj):
    ref = _get_ref(obj)
    return _trace.get_obj_expr(ref)


def type_of(obj: typing.Any) -> types.Type:
    return types.TypeRegistry.type_of(obj)


# def weave(obj: typing.Any) -> RuntimeConstNode:
#     return _weave_internal.make_const_node(type_of(obj), obj)  # type: ignore


def from_pandas(df):
    return _ops.pandas_to_awl(df)


#### Newer API below


def init(project_name: str) -> _graph_client.GraphClient:
    """Initialize weave tracking, logging to a wandb project.

    Logging is initialized globally, so you do not need to keep a reference
    to the return value of init.

    Following init, calls of weave.op() decorated functions will be logged
    to the specified project.

    Args:
        project_name: The name of the Weights & Biases project to log to.

    Returns:
        A Weave client.
    """
    return _weave_init.init_wandb(project_name).client


@contextlib.contextmanager
def wandb_client(project_name: str) -> typing.Iterator[_graph_client.GraphClient]:
    inited_client = _weave_init.init_wandb(project_name)
    try:
        yield inited_client.client
    finally:
        inited_client.reset()


# This is currently an internal interface. We'll expose something like it though ("offline" mode)
def init_local_client() -> _graph_client.GraphClient:
    return _weave_init.init_local().client


@contextlib.contextmanager
def local_client() -> typing.Iterator[_graph_client.GraphClient]:
    inited_client = _weave_init.init_local()
    try:
        yield inited_client.client
    finally:
        inited_client.reset()


def publish(obj: typing.Any, name: str) -> _ref_base.Ref:
    """Save and version a python object.

    If an object with name already exists, and the content hash of obj does
    not match the latest version of that object, a new version will be created.

    Args:
        obj: The object to save and version.
        name: The name to save the object under.

    Returns:
        A weave Ref to the saved object.
    """
    client = _graph_client_context.require_graph_client()

    ref = client.save_object(obj, name, "latest")

    print(f"Published {ref.type.root_type_class().name} to {ref.ui_url}")

    # Have to manually put the ref on the obj, this is supposed to happen at
    # a lower level, but we use a mapper in _direct_publish that I think changes
    # the object identity
    _ref_base._put_ref(obj, ref)
    return ref


def ref(location: str) -> _ref_base.Ref:
    """Construct a Ref to a Weave object.

    TODO: what happens if obj does not exist

    Args:
        location: A fully-qualified weave ref URI, or if weave.init() has been called, "name:version" or just "name" ("latest" will be used for version in this case).


    Returns:
        A weave Ref to the object.
    """
    if not "://" in location:
        client = _graph_client_context.get_graph_client()
        if not client:
            raise ValueError("Call weave.init() first, or pass a fully qualified uri")
        if "/" in location:
            raise ValueError("'/' not currently supported in short-form URI")
        if ":" not in location:
            name = location
            version = "latest"
        else:
            name, version = location.split(":")
        location = str(client.ref_uri(name, version, "obj"))

    return _ref_base.Ref.from_str(location)


def obj_ref(obj: typing.Any) -> typing.Optional[_ref_base.Ref]:
    return _ref_base.get_ref(obj)


def output_of(obj: typing.Any) -> typing.Optional[_run.Run]:
    client = _graph_client_context.require_graph_client()

    ref = obj_ref(obj)
    if ref is None:
        return ref

    return client.ref_output_of(ref)


def as_op_def(fn: typing.Callable) -> OpDef:
    """Given a @weave.op() decorated function, return its OpDef.

    @weave.op() decorated functions are instances of OpDef already, so this
    function should be a no-op at runtime. But you can use it to satisfy type checkers
    if you need to access OpDef attributes in a typesafe way.

    Args:
        fn: A weave.op() decorated function.

    Returns:
        The OpDef of the function.
    """
    if not isinstance(fn, OpDef):
        raise ValueError("fn must be a weave.op() decorated function")
    return fn


import contextlib


@contextlib.contextmanager
def attributes(attributes: typing.Dict[str, typing.Any]) -> typing.Iterator:
    cur_attributes = {**_monitor._attributes.get()}
    cur_attributes.update(attributes)

    token = _monitor._attributes.set(cur_attributes)
    try:
        yield
    finally:
        _monitor._attributes.reset(token)


def serve(
    model_ref: _artifact_wandb.WandbArtifactRef,
    method_name: typing.Optional[str] = None,
    auth_entity: typing.Optional[str] = None,
    port: int = 9996,
    thread: bool = False,
) -> typing.Optional[str]:
    import uvicorn
    from .serve_fastapi import object_method_app

    client = _graph_client_context.require_graph_client()
    if not isinstance(
        client, _graph_client_wandb_art_st.GraphClientWandbArtStreamTable
    ):
        raise ValueError("serve currently only supports wandb client")

    print(f"Serving {model_ref}")
    print(f"🥐 Server docs and playground at http://localhost:{port}/docs")
    print()
    os.environ["PROJECT_NAME"] = f"{client.entity_name}/{client.project_name}"
    os.environ["MODEL_REF"] = str(model_ref)

    wandb_api_ctx = _wandb_api.get_wandb_api_context()
    app = object_method_app(model_ref, method_name=method_name, auth_entity=auth_entity)
    trace_attrs = _monitor._attributes.get()

    def run():
        with _wandb_api.wandb_api_context(wandb_api_ctx):
            with attributes(trace_attrs):
                uvicorn.run(app, host="0.0.0.0", port=port)

    if _util.is_notebook():
        thread = True
    if thread:
        from threading import Thread

        t = Thread(target=run, daemon=True)
        t.start()
        time.sleep(1)
        return "http://localhost:%d" % port
    else:
        run()
    return None


__docspec__ = [init, publish, ref]



# Safe class for managing W&B Artifact downloads. Can be used
# as the core of a fast artifact downloader, or a safe multiplexed
# file manager in the Weave server.

import datetime
import json
import typing
import urllib
from aiohttp import BasicAuth
from requests.auth import HTTPBasicAuth

from wandb.sdk.lib import hashutil


from . import artifact_wandb
from . import errors
from . import engine_trace
from . import filesystem
from . import weave_http
from . import wandb_api
from . import environment as weave_env
from . import cache


from urllib import parse

tracer = engine_trace.tracer()  # type: ignore


def _file_path(uri: artifact_wandb.WeaveWBArtifactURI, md5_hex: str) -> str:
    if uri.path is None:
        raise errors.WeaveInternalError("Artifact URI has no path in call to file_path")
    path_parts = uri.path.split(".", 1)
    if len(path_parts) == 2:
        extension = "." + path_parts[1]
    else:
        extension = ""
    return f"wandb_file_manager/{uri.entity_name}/{uri.project_name}/{uri.name}/{md5_hex}{extension}"


def _local_path_and_download_url(
    art_uri: artifact_wandb.WeaveWBArtifactURI,
    manifest: artifact_wandb.WandbArtifactManifest,
) -> typing.Optional[typing.Tuple[str, str]]:
    path = art_uri.path
    if path is None:
        raise errors.WeaveInternalError(
            "Artifact URI has no path in call to _local_path_and_download_url"
        )
    file_name = path.split("/")[-1]
    manifest_entry = manifest.get_entry_by_path(path)
    if manifest_entry is None:
        return None
    md5_hex = hashutil.b64_to_hex_id(hashutil.B64MD5(manifest_entry["digest"]))
    base_url = weave_env.wandb_base_url()
    file_path = _file_path(art_uri, md5_hex)
    if manifest.storage_layout == artifact_wandb.WandbArtifactManifest.StorageLayout.V1:
        return file_path, "{}/artifacts/{}/{}/{}".format(
            base_url, art_uri.entity_name, md5_hex, urllib.parse.quote(file_name)
        )
    else:
        # TODO: storage_region
        storage_region = "default"
        return file_path, "{}/artifactsV2/{}/{}/{}/{}/{}".format(
            base_url,
            storage_region,
            art_uri.entity_name,
            urllib.parse.quote(manifest_entry.get("birthArtifactID", "")),  # type: ignore
            md5_hex,
            urllib.parse.quote(file_name),
        )


class WandbFileManagerAsync:
    def __init__(
        self,
        filesystem: filesystem.FilesystemAsync,
        http: weave_http.HttpAsync,
        wandb_api: wandb_api.WandbApiAsync,
    ) -> None:
        self.fs = filesystem
        self.http = http
        self.wandb_api = wandb_api
        self._manifests: cache.LruTimeWindowCache[
            str, typing.Optional[artifact_wandb.WandbArtifactManifest]
        ] = cache.LruTimeWindowCache(datetime.timedelta(minutes=5))

    def manifest_path(self, uri: artifact_wandb.WeaveWBArtifactURI) -> str:
        assert uri.version is not None
        return f"wandb_file_manager/{uri.entity_name}/{uri.project_name}/{uri.name}/manifest-{uri.version}.json"

    async def _manifest(
        self, art_uri: artifact_wandb.WeaveWBArtifactURI, manifest_path: str
    ) -> typing.Optional[artifact_wandb.WandbArtifactManifest]:
        if art_uri.version is None:
            raise errors.WeaveInternalError(
                'Artifact URI has no version: "%s"' % art_uri
            )
        # Check if on disk
        try:
            async with self.fs.open_read(manifest_path) as f:
                return artifact_wandb.WandbArtifactManifest(json.loads(await f.read()))
        except FileNotFoundError:
            pass
        # Download
        manifest_url = await self.wandb_api.artifact_manifest_url(
            art_uri.entity_name,
            art_uri.project_name,
            art_uri.name + ":" + art_uri.version,
        )
        if manifest_url is None:
            return None
        await self.http.download_file(
            manifest_url,
            manifest_path,
        )
        async with self.fs.open_read(manifest_path) as f:
            return artifact_wandb.WandbArtifactManifest(json.loads(await f.read()))

    async def manifest(
        self, art_uri: artifact_wandb.WeaveWBArtifactURI
    ) -> typing.Optional[artifact_wandb.WandbArtifactManifest]:
        with tracer.trace("wandb_file_manager.manifest") as span:
            assert art_uri.version is not None
            manifest_path = self.manifest_path(art_uri)
            manifest = self._manifests.get(manifest_path)
            if not isinstance(manifest, cache.LruTimeWindowCache.NotFound):
                return manifest
            manifest = await self._manifest(art_uri, manifest_path)
            self._manifests.set(manifest_path, manifest)
            return manifest

    async def local_path_and_download_url(
        self, art_uri: artifact_wandb.WeaveWBArtifactURI
    ) -> typing.Optional[typing.Tuple[str, str]]:
        path = art_uri.path
        if path is None:
            raise errors.WeaveInternalError(
                "Artifact URI has no path in call to local_path_and_download_url"
            )
        manifest = await self.manifest(art_uri)
        if manifest is None:
            return None
        return _local_path_and_download_url(art_uri, manifest)

    async def ensure_file(
        self, art_uri: artifact_wandb.WeaveWBArtifactURI
    ) -> typing.Optional[str]:
        path = art_uri.path
        if path is None:
            raise errors.WeaveInternalError(
                "Artifact URI has no path in call to ensure_file"
            )
        with tracer.trace("wandb_file_manager.ensure_file") as span:
            span.set_tag("uri", str(art_uri))
            res = await self.local_path_and_download_url(art_uri)
            if res is None:
                return None
            file_path, download_url = res
            if await self.fs.exists(file_path):
                return file_path
            wandb_api_context = wandb_api.get_wandb_api_context()
            headers = None
            cookies = None
            auth = None
            if wandb_api_context is not None:
                headers = wandb_api_context.headers
                cookies = wandb_api_context.cookies
                if wandb_api_context.api_key is not None:
                    auth = BasicAuth("api", wandb_api_context.api_key)
            await self.http.download_file(
                download_url,
                file_path,
                headers=headers,
                cookies=cookies,
                auth=auth,
            )
            return file_path

    async def ensure_file_downloaded(
        self,
        download_url: str,
    ) -> typing.Optional[str]:
        """Ensures a history parquet file from an http/https URI."""
        schema, netloc, path, _, _, _ = parse.urlparse(download_url)
        path = "/".join([netloc, path.lstrip("/")])
        if schema not in ("http", "https"):
            raise errors.WeaveInternalError(
                "ensure_file_downloaded only supports http/https URIs"
            )
        if path is None:
            raise errors.WeaveInternalError(
                "Artifact URI has no path in call to ensure_file"
            )
        with tracer.trace("wandb_file_manager.ensure_file_downloaded") as span:
            span.set_tag("download_url", str(download_url))
            file_path = f"wandb_file_manager/{path}"
            if await self.fs.exists(file_path):
                return file_path
            wandb_api_context = wandb_api.get_wandb_api_context()
            headers = None
            cookies = None
            auth = None
            if wandb_api_context is not None:
                headers = wandb_api_context.headers
                cookies = wandb_api_context.cookies
                if wandb_api_context.api_key is not None:
                    auth = BasicAuth("api", wandb_api_context.api_key)
            await self.http.download_file(
                download_url,
                file_path,
                headers=headers,
                cookies=cookies,
                auth=auth,
            )
            return file_path

    async def direct_url(
        self, art_uri: artifact_wandb.WeaveWBArtifactURI
    ) -> typing.Optional[str]:
        path = art_uri.path
        if path is None:
            raise errors.WeaveInternalError(
                "Artifact URI has no path in call to ensure_file"
            )
        with tracer.trace("wandb_file_manager.direct_url") as span:
            span.set_tag("uri", str(art_uri))
            res = await self.local_path_and_download_url(art_uri)
            if res is None:
                return None
            return res[1]


class WandbFileManager:
    def __init__(
        self,
        filesystem: filesystem.Filesystem,
        http: weave_http.Http,
        wandb_api: wandb_api.WandbApi,
    ) -> None:
        self.fs = filesystem
        self.http = http
        self.wandb_api = wandb_api
        self._manifests: cache.LruTimeWindowCache[
            str, typing.Optional[artifact_wandb.WandbArtifactManifest]
        ] = cache.LruTimeWindowCache(datetime.timedelta(minutes=5))

    def manifest_path(self, uri: artifact_wandb.WeaveWBArtifactURI) -> str:
        assert uri.version is not None
        return f"wandb_file_manager/{uri.entity_name}/{uri.project_name}/{uri.name}/manifest-{uri.version}.json"

    def _manifest(
        self, art_uri: artifact_wandb.WeaveWBArtifactURI, manifest_path: str
    ) -> typing.Optional[artifact_wandb.WandbArtifactManifest]:
        if art_uri.version is None:
            raise errors.WeaveInternalError(
                'Artifact URI has no version: "%s"' % art_uri
            )
        # Check if on disk
        try:
            with self.fs.open_read(manifest_path) as f:
                return artifact_wandb.WandbArtifactManifest(json.loads(f.read()))
        except FileNotFoundError:
            pass
        # Download
        manifest_url = self.wandb_api.artifact_manifest_url(
            art_uri.entity_name,
            art_uri.project_name,
            art_uri.name + ":" + art_uri.version,
        )
        if manifest_url is None:
            return None
        self.http.download_file(
            manifest_url,
            manifest_path,
        )
        with self.fs.open_read(manifest_path) as f:
            return artifact_wandb.WandbArtifactManifest(json.loads(f.read()))

    def manifest(
        self, art_uri: artifact_wandb.WeaveWBArtifactURI
    ) -> typing.Optional[artifact_wandb.WandbArtifactManifest]:
        with tracer.trace("wandb_file_manager.manifest") as span:
            assert art_uri.version is not None
            manifest_path = self.manifest_path(art_uri)
            manifest = self._manifests.get(manifest_path)
            if not isinstance(manifest, cache.LruTimeWindowCache.NotFound):
                return manifest
            manifest = self._manifest(art_uri, manifest_path)
            self._manifests.set(manifest_path, manifest)
            return manifest

    def local_path_and_download_url(
        self, art_uri: artifact_wandb.WeaveWBArtifactURI
    ) -> typing.Optional[typing.Tuple[str, str]]:
        path = art_uri.path
        if path is None:
            raise errors.WeaveInternalError(
                "Artifact URI has no path in call to local_path_and_download_url"
            )
        manifest = self.manifest(art_uri)
        if manifest is None:
            return None
        return _local_path_and_download_url(art_uri, manifest)

    def ensure_file_downloaded(
        self,
        download_url: str,
    ) -> typing.Optional[str]:
        """Ensures a history parquet file from an http/https URI."""
        schema, netloc, path, _, _, _ = parse.urlparse(download_url)
        path = "/".join([netloc, path.lstrip("/")])
        if schema not in ("http", "https"):
            raise errors.WeaveInternalError(
                "ensure_file_downloaded only supports http/https URIs"
            )
        if path is None:
            raise errors.WeaveInternalError(
                "Artifact URI has no path in call to ensure_file"
            )
        with tracer.trace("wandb_file_manager.ensure_file_downloaded") as span:
            span.set_tag("download_url", str(download_url))
            file_path = f"wandb_file_manager/{path}"
            if self.fs.exists(file_path):
                return file_path
            wandb_api_context = wandb_api.get_wandb_api_context()
            headers = None
            cookies = None
            auth = None
            if wandb_api_context is not None:
                headers = wandb_api_context.headers
                cookies = wandb_api_context.cookies
                if wandb_api_context.api_key is not None:
                    auth = HTTPBasicAuth("api", wandb_api_context.api_key)
            self.http.download_file(
                download_url, file_path, headers=headers, cookies=cookies, auth=auth
            )
            return file_path

    def ensure_file(
        self, art_uri: artifact_wandb.WeaveWBArtifactURI
    ) -> typing.Optional[str]:
        path = art_uri.path
        if path is None:
            raise errors.WeaveInternalError(
                "Artifact URI has no path in call to ensure_file"
            )
        with tracer.trace("wandb_file_manager.ensure_file") as span:
            span.set_tag("uri", str(art_uri))
            res = self.local_path_and_download_url(art_uri)
            if res is None:
                return None
            file_path, download_url = res
            if self.fs.exists(file_path):
                return file_path
            wandb_api_context = wandb_api.get_wandb_api_context()
            headers = None
            cookies = None
            auth = None
            if wandb_api_context is not None:
                headers = wandb_api_context.headers
                cookies = wandb_api_context.cookies
                if wandb_api_context.api_key is not None:
                    auth = HTTPBasicAuth("api", wandb_api_context.api_key)
            self.http.download_file(
                download_url, file_path, headers=headers, cookies=cookies, auth=auth
            )
            return file_path

    def direct_url(
        self, art_uri: artifact_wandb.WeaveWBArtifactURI
    ) -> typing.Optional[str]:
        path = art_uri.path
        if path is None:
            raise errors.WeaveInternalError(
                "Artifact URI has no path in call to ensure_file"
            )
        with tracer.trace("wandb_file_manager.direct_url") as span:
            span.set_tag("uri", str(art_uri))
            res = self.local_path_and_download_url(art_uri)
            if res is None:
                return None
            return res[1]



import typing
from . import graph_client
from . import graph_client_local
from . import graph_client_wandb_art_st
from . import context_state
from . import errors
from . import autopatch


class InitializedClient:
    def __init__(self, client: graph_client.GraphClient):
        self.client = client
        self.graph_client_token = context_state._graph_client.set(client)
        self.ref_tracking_token = context_state._ref_tracking_enabled.set(True)
        self.eager_mode_token = context_state._eager_mode.set(True)
        self.serverless_io_service_token = context_state._serverless_io_service.set(
            True
        )

    def reset(self) -> None:
        context_state._graph_client.reset(self.graph_client_token)
        context_state._ref_tracking_enabled.reset(self.ref_tracking_token)
        context_state._eager_mode.reset(self.eager_mode_token)
        context_state._serverless_io_service.reset(self.serverless_io_service_token)


def init_wandb(project_name: str) -> InitializedClient:
    from . import wandb_api

    fields = project_name.split("/")
    if len(fields) == 1:
        api = wandb_api.get_wandb_api_sync()
        try:
            entity_name = api.default_entity_name()
        except AttributeError:
            raise errors.WeaveWandbAuthenticationException(
                'weave init requires wandb. Run "wandb login"'
            )
        project_name = fields[0]
    elif len(fields) == 2:
        entity_name, project_name = fields
    else:
        raise ValueError(
            'project_name must be of the form "<project_name>" or "<entity_name>/<project_name>"'
        )
    if not entity_name:
        raise ValueError("entity_name must be non-empty")
    client = graph_client_wandb_art_st.GraphClientWandbArtStreamTable(
        entity_name, project_name
    )

    init_client = InitializedClient(client)

    # autopatching is only supporte for the wandb client, because OpenAI calls are not
    # logged in local mode currently. When that's fixed, this autopatch call can be
    # moved to InitializedClient.__init__
    autopatch.autopatch()

    return init_client


def init_local() -> InitializedClient:
    client = graph_client_local.GraphClientLocal()
    return InitializedClient(client)



import typing
from . import panel
from . import graph
from . import weave_types
from . import storage


def make_node(v: typing.Any) -> graph.Node:
    """Business logic for how values passed to panels are converted to json."""
    from . import ops

    if isinstance(v, graph.Node):
        return v

    node_type = weave_types.TypeRegistry.type_of(v)
    if weave_types.is_json_compatible(node_type):
        return graph.ConstNode(node_type, v)

    # Otherwise
    ref = storage.save(v)
    return ops.get(str(ref))  # type: ignore


def child_item(v):
    if isinstance(v, panel.Panel):
        return v
    else:
        return make_node(v)



# This file contains a set of routines for converting the return type of a GQL query to a weave type.
# It works by parsing the GQL query and then recursively walking the AST, converting each node to a weave type.
# This depends on having a GraphQL schema available, which is currently loaded from a file. The load
# happens when this file is imported and used throughout the duration of the program.

from graphql import (
    parse,
    get_operation_root_type,
    GraphQLObjectType,
    GraphQLList,
    GraphQLNonNull,
    GraphQLInterfaceType,
    FieldNode,
    OperationDefinitionNode,
)
import graphql
import typing

from . import errors
from . import weave_types as types
from . import gql_schema


def get_outermost_alias(query_str: str) -> str:
    gql_doc = graphql.language.parse(f"query innerquery {{ {query_str} }}")
    root_operation = gql_doc.definitions[0]
    if not isinstance(root_operation, graphql.language.ast.OperationDefinitionNode):
        raise errors.WeaveInternalError("Only operation definitions are supported.")
    if len(root_operation.selection_set.selections) != 1:
        # NOTE: if we ever need a root op to have multiple root selections, we
        # can easily loosen this restriction and just return a list of aliases
        raise errors.WeaveInternalError("Only one root selection is supported")
    inner_selection = root_operation.selection_set.selections[0]
    if not isinstance(inner_selection, graphql.language.ast.FieldNode):
        raise errors.WeaveInternalError("Only field selections are supported")
    if inner_selection.alias is not None:
        return inner_selection.alias.value
    return inner_selection.name.value


def gql_type_to_weave_type(
    gql_type: graphql.GraphQLType,
    selection_set: typing.Optional[graphql.SelectionSetNode],
) -> types.Type:
    if (
        isinstance(gql_type, (GraphQLObjectType, GraphQLInterfaceType))
        and selection_set
    ):
        property_types: dict[str, types.Type] = {}
        selections: list[graphql.SelectionNode] = []

        # Handle inline fragments (i.e., ... on Foo)
        for selection in selection_set.selections:
            if (
                isinstance(selection, graphql.InlineFragmentNode)
                and selection.type_condition.name.value == gql_type.name
            ):
                selections.extend(selection.selection_set.selections)
            elif isinstance(selection, FieldNode):
                selections.append(selection)

        for selection in selections:
            if not isinstance(selection, FieldNode):
                raise ValueError(
                    f"Selections must be fields, got {selection.__class__.__name__}"
                )
            key = selection.alias.value if selection.alias else selection.name.value
            if key == "__typename":
                # __typename does not appear explicitly in the schema, but all types have it
                # it just returns a string, so treat that case here.
                property_types[key] = types.Const(types.String(), gql_type.name)
            else:
                property_types[key] = gql_type_to_weave_type(
                    gql_type.fields[selection.name.value].type, selection.selection_set
                )
        return types.TypedDict(property_types)

    elif isinstance(gql_type, GraphQLList):
        # Coerce JSON[] to JSON - JSON[] is bad for performance
        of_type = gql_type.of_type
        if isinstance(of_type, GraphQLNonNull):
            if of_type.of_type.name == "JSON":
                return types.String()
        elif of_type.name == "JSON":
            return types.String()
        return types.List(gql_type_to_weave_type(of_type, selection_set))
    elif isinstance(gql_type, GraphQLNonNull):
        return types.non_none(gql_type_to_weave_type(gql_type.of_type, selection_set))
    elif isinstance(gql_type, graphql.GraphQLUnionType):
        return types.union(
            *[gql_type_to_weave_type(t, selection_set) for t in gql_type.types]
        )
    elif isinstance(gql_type, graphql.GraphQLScalarType):
        t: types.Type
        if gql_type.name in [
            "String",
            "ID",
            "JSONString",
        ]:
            t = types.String()
        elif gql_type.name in ["Int", "Int64"]:
            t = types.Int()
        elif gql_type.name == "Float":
            t = types.Float()
        elif gql_type.name == "Boolean":
            t = types.Boolean()
        elif gql_type.name == "JSON":
            t = types.String()
        elif gql_type.name == "DateTime":
            t = types.Timestamp()
        elif gql_type.name == "Duration":
            t = types.Number()  # duration is number of seconds
        else:
            raise ValueError(f"Unknown scalar type {gql_type.name}")

        return types.optional(t)

    raise ValueError(f"Unknown type {gql_type}")


def get_query_weave_type(query: str) -> types.Type:
    """
    Converts a given GraphQL query to a corresponding Weave type.

    This function parses the provided query, determines its root operation type from the schema,
    and then converts that root operation type into a Weave type using the gql_type_to_weave_type() function.

    Parameters:
    query (str): The GraphQL query string to be converted.

    Returns:
    types.Type: The corresponding Weave type.

    Raises:
    ValueError: If no operation is found in the query.

    Note:
    It prints the query string to the console. In a production setting, consider using logging for this instead.
    """
    document = parse(query)
    for definition in document.definitions:
        if isinstance(definition, OperationDefinitionNode):
            schema = gql_schema.gql_schema()
            root_operation_type = get_operation_root_type(schema, definition)
            return gql_type_to_weave_type(root_operation_type, definition.selection_set)
    raise ValueError("No operation found in query")



import typing

import weave
from ..ref_base import Ref
from .dataset import Dataset


def publish_huggingface_dataset(
    hf_dataset_name: str,
    config_name: typing.Optional[str],
    split_name: str,
    sample_size: int,
    seed: int,
) -> Ref:
    import random
    from datasets import load_dataset

    if config_name is None:
        config_name = "default"

    dataset = load_dataset(hf_dataset_name, config_name)
    split = dataset[split_name]
    random.seed(seed)
    indexes = random.sample(range(len(split)), sample_size)
    sample = split.select(indexes)
    return weave.publish(
        Dataset(list(sample)),
        name=f"{hf_dataset_name}-{split_name}-{sample_size}-{seed}",
    )



# TODO: types are all messed up after panel_composition branch
import copy
import typing
import weave
from weave import codifiable_value_mixin
from weave import codify

from .. import panel
from . import table_state
from . import panel_table
import dataclasses

from .. import graph
from .. import errors
from .. import weave_internal
from .. import weave_types as types
from ..ops_primitives import list_, dict as dict_, boolean


@weave.type()
class DimConfig:
    x: str
    y: str
    color: str
    label: str
    tooltip: str
    pointSize: str
    pointShape: str
    y2: str


DimName = typing.Literal[
    "x",
    "y",
    "color",
    "label",
    "tooltip",
    "pointSize",
    "pointShape",
    "y2",
]
MarkOption = typing.Optional[typing.Literal["area", "line", "point", "bar", "boxplot"]]
PointShapeOption = typing.Literal[
    "series",
    "circle",
    "square",
    "cross",
    "diamond",
    "triangle-up",
    "triangle-down",
    "triangle-right",
    "triangle-left",
]

LabelOption = typing.Literal["series"]
LineStyleOption = typing.Literal[
    "solid",
    "dashed",
    "dotted",
    "dot-dashed",
    "short-dashed",
    "series",
]

LAZY_PATHS = [
    "signals.domain.x",
    "signals.domain.y",
]

LAZY_PATH_TYPES = {
    "signals.domain.x": weave.types.optional(weave.types.List(weave.types.Any())),
    "signals.domain.y": weave.types.optional(weave.types.List(weave.types.Any())),
}


SelectFunction = typing.Union[graph.Node, typing.Callable[[typing.Any], typing.Any]]


@weave.type()
class PlotConstants:
    mark: typing.Optional[MarkOption] = None
    pointShape: PointShapeOption = "circle"
    label: LabelOption = "series"
    lineStyle: LineStyleOption = "solid"


class PlotConstantsInputDict(typing.TypedDict):
    # TODO(DG): replace these typing.Optionals with typing.NotRequired (https://peps.python.org/pep-0655/#usage-in-python-3-11)
    # when we drop support for Python 3.10 and python 3.11 is released

    # this will mimic the behavior of ? in typescript
    mark: typing.Optional[MarkOption]
    pointShape: typing.Optional[PointShapeOption]
    label: typing.Optional[LabelOption]
    lineStyle: typing.Optional[LineStyleOption]


DropdownWithExpressionMode = typing.Union[
    typing.Literal["dropdown"], typing.Literal["expression"]
]


@weave.type()
class PlotUIState:
    pointShape: DropdownWithExpressionMode
    label: DropdownWithExpressionMode


def default_axis():
    return AxisSetting(noLabels=False, noTitle=False, noTicks=False, scale=None)


@weave.type()
class Series:
    table: table_state.TableState
    dims: DimConfig
    constants: PlotConstants
    uiState: PlotUIState

    @property
    def y_expr_str(self):
        y_col_id = self.dims.y
        y_fn = self.table.columnSelectFunctions[y_col_id]
        if isinstance(y_fn, graph.Node):
            return graph.node_expr_str(y_fn)
        return None

    def clone(self):
        series = Series.__new__(Series)
        for attr in ["table", "dims", "constants", "uiState"]:
            value = getattr(self, attr)
            if hasattr(value, "clone"):
                setattr(series, attr, value.clone())
            else:
                setattr(series, attr, copy.deepcopy(value))
        return series

    def __init__(
        self,
        input_node: typing.Optional[graph.Node] = None,
        table: typing.Optional[table_state.TableState] = None,
        dims: typing.Optional[DimConfig] = None,
        constants: typing.Optional[PlotConstants] = None,
        select_functions: typing.Optional[typing.Dict[DimName, SelectFunction]] = None,
        groupby_dims: typing.Optional[list[DimName]] = None,
    ):
        if input_node is not None:
            table = table_state.TableState(input_node)
        if table is None:
            raise ValueError("table missing")

        if dims is None:
            _dims = {}
            for dim in dataclasses.fields(DimConfig):
                _dims[dim.name] = table.add_column(lambda row: graph.VoidNode())
            dims = DimConfig(**_dims)
        if constants is None:
            constants = PlotConstants()

        groupby_dims = groupby_dims or []

        if select_functions:
            for dimname in groupby_dims:
                if dimname not in select_functions:
                    raise ValueError(
                        f"select_functions must contain a function for groupby dim {dimname}"
                    )
                table.enable_groupby(getattr(dims, dimname))
                if dimname == "label":
                    table.enable_groupby(dims.color)

                table.update_col(getattr(dims, dimname), select_functions[dimname])

            for dimname in select_functions:
                if dimname not in groupby_dims:
                    table.update_col(getattr(dims, dimname), select_functions[dimname])

        uiState = PlotUIState(pointShape="expression", label="expression")

        self.table = table
        self.dims = dims
        self.constants = constants
        self.uiState = uiState

    @classmethod
    def from_config(cls, **config) -> "Series":
        series = Series.__new__(Series)
        for attr in ["table", "dims", "constants", "uiState"]:
            value = config.get(attr, None)
            if value is None:
                raise ValueError(
                    f'Config must contain "{attr}" to initialize new series.'
                )
            setattr(series, attr, value)
        return series

    @property
    def table_query(self):
        return self.table


def make_set(dim_name: str) -> typing.Callable[[Series, typing.Any], None]:
    def set(self: Series, expr) -> None:
        col_id: str = getattr(self.dims, dim_name)
        self.table.update_col(col_id, expr)

    return set


def make_group(dim_name: str) -> typing.Callable[[Series], None]:
    def group(self: Series) -> None:
        col_id: str = getattr(self.dims, dim_name)
        self.table.enable_groupby(col_id)

    return group


def make_get(dim_name: str) -> typing.Callable[[Series], graph.Node]:
    def get(self: Series) -> graph.Node:
        col_id: str = getattr(self.dims, dim_name)
        return self.table.columnSelectFunctions[col_id]

    return get


def make_set_constant(
    const_name: str,
) -> typing.Callable[[Series, typing.Any], None]:
    def set(self: Series, value) -> None:
        setattr(self.constants, const_name, value)

    return set


def make_get_constant(const_name: str) -> typing.Callable[[Series], typing.Any]:
    def get(self: Series) -> typing.Any:
        return getattr(self.constants, const_name)

    return get


# install setters and groupers
for dim in dataclasses.fields(DimConfig):
    setattr(Series, f"set_{dim.name}", make_set(dim.name))
    setattr(Series, f"groupby_{dim.name}", make_group(dim.name))
    setattr(Series, f"get_{dim.name}", make_get(dim.name))

for dim in dataclasses.fields(PlotConstants):
    setattr(Series, f"set_{dim.name}_constant", make_set_constant(dim.name))
    setattr(Series, f"get_{dim.name}_constant", make_get_constant(dim.name))


ScaleType = typing.Literal["linear", "log"]


@weave.type()
class AxisScale:
    # need to use scaleType instead of `type` here because
    # `type` serializes to the same key as the
    # `type` used by ObjectType
    scaleType: typing.Optional[ScaleType] = "linear"
    range: typing.Optional[
        dict[typing.Literal["field"], typing.Callable[[str], str]]
    ] = None
    base: typing.Optional[float] = None  # for log scale


@weave.type()
class AxisSetting:
    noLabels: bool = False
    noTitle: bool = False
    noTicks: bool = False
    title: typing.Optional[str] = None
    scale: typing.Optional[AxisScale] = None


@weave.type()
class AxisSettings:
    x: AxisSetting
    y: AxisSetting
    color: AxisSetting
    # todo: figure out how to reproduce [key: string]: LegendSetting in python type system


@weave.type()
class LegendSetting:
    noLegend: bool = False


@weave.type()
class LegendSettings:
    x: LegendSetting = dataclasses.field(default_factory=LegendSetting)
    y: LegendSetting = dataclasses.field(default_factory=LegendSetting)
    color: LegendSetting = dataclasses.field(default_factory=LegendSetting)
    pointShape: LegendSetting = dataclasses.field(default_factory=LegendSetting)
    pointSize: LegendSetting = dataclasses.field(default_factory=LegendSetting)
    lineStyle: LegendSetting = dataclasses.field(default_factory=LegendSetting)


@weave.type()
class ConfigOptionsExpanded:
    x: bool = False
    y: bool = False
    label: bool = False
    tooltip: bool = False
    mark: bool = False


DiscreteSelection = typing.List[str]
ContinuousSelection = typing.Tuple[float, float]
Selection = typing.Optional[typing.Union[ContinuousSelection, DiscreteSelection]]

LazySelection = typing.Union[Selection, SelectFunction]


# TODO: split this into 2 - one for lazy, one eager
@weave.type()
class AxisSelections:
    x: typing.Optional[Selection] = None
    y: typing.Optional[Selection] = None


@weave.type()
class LazyAxisSelections:
    x: SelectFunction = graph.ConstNode(weave.types.optional(weave.types.Any()), None)
    y: SelectFunction = graph.ConstNode(weave.types.optional(weave.types.Any()), None)


@weave.type()
class Signals:
    domain: LazyAxisSelections
    selection: AxisSelections


@weave.type("plotConfig")
class PlotConfig:
    series: typing.List[Series]
    axisSettings: typing.Optional[AxisSettings]
    legendSettings: typing.Optional[LegendSettings]
    configOptionsExpanded: typing.Optional[ConfigOptionsExpanded]
    signals: Signals
    configVersion: int = 15


def set_through_array(
    obj: typing.Any, path: typing.List[str], value: typing.Any
) -> None:
    """
    Recursively sets an attribute, key or index specified by a path in a nested object structure to a given value
    (consisting of dictionaries, lists or custom objects). The path is a list of strings, with the special value '#'
    used to denote all items in a list.

    Parameters:
    obj (Any): The object that will be traversed and modified.
    path (List[str]): The list of strings, where each string is a key or attribute name in the path to the location
    in the nested object structure to be modified. If a string is '#', it indicates that all items in the list at
    that level should be modified.
    value (Any): The value to set at the location specified by the path.

    Raises:
    ValueError: If the path cannot be followed in the object, it raises a ValueError.

    Returns:
    None: The function modifies the object in-place and does not return any value.
    """
    if len(path) == 0:
        return

    first, *rest = path

    if first == "#":
        if isinstance(obj, list):
            for i, item in enumerate(obj):
                set_through_array(
                    item, rest, (value[i] if isinstance(value, list) else value)
                )
    elif len(rest) == 0:
        if isinstance(obj, dict):
            obj[first] = value
        else:
            setattr(obj, first, value)
    elif hasattr(obj, first):
        set_through_array(getattr(obj, first), rest, value)
    else:
        raise ValueError(f"Could not set {path} in {obj} to {value}")


def get_through_array(
    obj: typing.Any, path: typing.List[str], coerce_list_of_nodes_to_node=False
) -> typing.Any:
    """
    Recursively retrieves a value from an attribute, key or index specified by a path in a nested object structure
    (consisting of dictionaries, lists or custom objects). The path is a list of strings, with the special value '#'
    used to denote all items in a list.

    Parameters:
    obj (Any): The object that will be traversed.
    path (List[str]): The list of strings, where each string is a key or attribute name in the path to the location
    in the nested object structure to retrieve. If a string is '#', it indicates that values from all items in the
    list at that level should be retrieved.
    coerce_list_of_nodes_to_node (bool): This flag controls how lists of 'graph.Node' objects are handled. If set to
    True, lists of 'graph.Node' objects are turned into a single 'list_.make_list' object with nodes as its items.

    Raises:
    ValueError: If the path cannot be followed in the object, it raises a ValueError.

    Returns:
    Any: The value or list of values retrieved from the object at the location specified by the path.
    """
    if len(path) == 0:
        return obj

    first, *rest = path

    if first == "#":
        if isinstance(obj, list):
            result = [
                get_through_array(
                    item,
                    rest,
                    coerce_list_of_nodes_to_node=coerce_list_of_nodes_to_node,
                )
                for item in obj
            ]
            if coerce_list_of_nodes_to_node and all(
                isinstance(item, graph.Node) for item in result
            ):
                result = list_.make_list({i: item for i, item in enumerate(result)})
            return result
        else:
            raise ValueError(f"Could not get {path} in {obj}: not a list")
    elif hasattr(obj, first):
        return get_through_array(
            getattr(obj, first),
            rest,
            coerce_list_of_nodes_to_node=coerce_list_of_nodes_to_node,
        )
    elif len(rest) == 0:
        if isinstance(obj, dict):
            return obj[first]
        else:
            return getattr(obj, first, None)
    else:
        raise ValueError(f"Could not get {path} in {obj}")


def ensure_node(
    obj: typing.Any,
    path: typing.List[str],
    type: typing.Optional[weave.types.Type] = None,
) -> None:
    value = get_through_array(obj, path)
    if isinstance(value, list):
        if not all(item == value[0] for item in value):
            raise ValueError(f"Cannot ensure node for {path}: not all values are equal")
        value = value[0]
    if not isinstance(value, graph.Node):
        if type is None:
            type = weave.types.TypeRegistry.type_of(value)
        set_through_array(obj, path, graph.ConstNode(type, value))


@weave.type("plot")
class Plot(panel.Panel, codifiable_value_mixin.CodifiableValueMixin):
    id = "plot"
    config: typing.Optional[PlotConfig] = None

    def __init__(
        self,
        input_node=None,
        vars=None,
        config: typing.Optional[PlotConfig] = None,
        constants: typing.Optional[PlotConstants] = None,
        mark: typing.Optional[MarkOption] = None,
        x: typing.Optional[SelectFunction] = None,
        y: typing.Optional[SelectFunction] = None,
        color: typing.Optional[SelectFunction] = None,
        label: typing.Optional[SelectFunction] = None,
        groupby_dims: typing.Optional[typing.List[DimName]] = None,
        tooltip: typing.Optional[SelectFunction] = None,
        pointShape: typing.Optional[SelectFunction] = None,
        lineStyle: typing.Optional[SelectFunction] = None,
        y2: typing.Optional[SelectFunction] = None,
        series: typing.Optional[typing.Union[Series, typing.List[Series]]] = None,
        no_axes: bool = False,
        no_legend: bool = False,
        x_title: typing.Optional[str] = None,
        y_title: typing.Optional[str] = None,
        x_axis_type: typing.Optional[ScaleType] = None,
        y_axis_type: typing.Optional[ScaleType] = None,
        color_title: typing.Optional[str] = None,
        domain_x: typing.Optional[SelectFunction] = None,
        domain_y: typing.Optional[SelectFunction] = None,
    ):
        super().__init__(input_node=input_node, vars=vars)
        self.config = config

        if self.config is None:
            constants = PlotConstants(mark=mark)

            select_functions: typing.Optional[dict[DimName, SelectFunction]] = None
            for field, maybe_dim in zip(
                dataclasses.fields(DimConfig),
                [x, y, color, label, tooltip, pointShape, lineStyle, y2],
            ):
                if maybe_dim is not None:
                    if select_functions is None:
                        select_functions = {}
                    select_functions[typing.cast(DimName, field.name)] = maybe_dim

            config_series: typing.List[Series]
            if series is not None:
                if isinstance(series, Series):
                    config_series = [series]
                else:
                    config_series = series
            else:
                config_series = [
                    Series(
                        typing.cast(graph.Node, self.input_node),
                        constants=constants,
                        select_functions=select_functions,
                        groupby_dims=groupby_dims,
                    )
                ]

            # Just moved these down so all series creation stuff is clustered
            # in the top.
            signals = Signals(domain=LazyAxisSelections(), selection=AxisSelections())
            if domain_x is not None:
                signals.domain.x = domain_x

            if domain_y is not None:
                signals.domain.y = domain_y

            config_axisSettings = AxisSettings(
                x=default_axis(), y=default_axis(), color=default_axis()
            )

            config_axisSettings.x.title = x_title
            config_axisSettings.y.title = y_title
            config_axisSettings.color.title = color_title

            if x_axis_type is not None:
                config_axisSettings.x.scale = AxisScale(scaleType=x_axis_type)
            if y_axis_type is not None:
                config_axisSettings.y.scale = AxisScale(scaleType=y_axis_type)

            config_legendSettings = LegendSettings(
                x=LegendSetting(),
                y=LegendSetting(),
                color=LegendSetting(),
                pointShape=LegendSetting(),
                lineStyle=LegendSetting(),
                pointSize=LegendSetting(),
            )

            self.config = PlotConfig(
                series=config_series,
                axisSettings=config_axisSettings,
                legendSettings=config_legendSettings,
                configOptionsExpanded=ConfigOptionsExpanded(),
                signals=signals,
            )

            for path in LAZY_PATHS:
                ensure_node(self.config, path.split("."), type=LAZY_PATH_TYPES[path])

            if no_axes:
                self.set_no_axes()

            if no_legend:
                self.set_no_legend()
        else:
            if self.config.series:
                for series in self.config.series:
                    series.table.input_node = self.input_node

    def add_series(self, series: Series):
        if self.config is None:
            raise errors.WeaveInternalError("config is None")
        self.config.series.append(series)

    def set_no_axes(self):
        if self.config is None:
            raise errors.WeaveInternalError("config is None")

        if self.config.axisSettings is None:
            self.config.axisSettings = AxisSettings(
                x=default_axis(), y=default_axis(), color=default_axis()
            )

        self.config.axisSettings.x = AxisSetting(
            noLabels=True,
            noTitle=True,
            noTicks=True,
            scale=None,
        )
        self.config.axisSettings.y = AxisSetting(
            noLabels=True,
            noTitle=True,
            noTicks=True,
            scale=None,
        )

    def set_no_legend(self):
        if self.config is None:
            raise errors.WeaveInternalError("config is None")

        self.config.legendSettings = LegendSettings(
            x=LegendSetting(True),
            y=LegendSetting(True),
            color=LegendSetting(True),
            pointShape=LegendSetting(),
            lineStyle=LegendSetting(),
            pointSize=LegendSetting(),
        )

    def to_code(self) -> typing.Optional[str]:
        field_vals: list[tuple[str, str]] = []
        default_signals = Signals(
            domain=LazyAxisSelections(),
            selection=AxisSelections(x=None, y=None),
        )
        default_axisSettings = AxisSettings(
            x=default_axis(), y=default_axis(), color=default_axis()
        )

        default_legendSettings = LegendSettings(
            x=LegendSetting(),
            y=LegendSetting(),
            color=LegendSetting(),
            pointShape=LegendSetting(),
            lineStyle=LegendSetting(),
            pointSize=LegendSetting(),
        )
        default_configOptionsExpanded = ConfigOptionsExpanded()

        # These are total hacks and should be removed once we have a proper way
        # to ensure UI produces the correct types
        default_axisSettings_asDict: dict = {"x": {}, "y": {}, "color": {}}
        default_legendSettings_asDict: dict = {"color": {}}
        default_configOptionsExpanded_asDict: dict = {
            "x": False,
            "y": False,
            "label": False,
            "tooltip": False,
            "mark": False,
        }

        if self.config is not None:
            if self.config.signals != default_signals:
                return None
            if (
                self.config.axisSettings != default_axisSettings
                and self.config.axisSettings != default_axisSettings_asDict
            ):
                return None
            if (
                self.config.legendSettings != default_legendSettings
                and self.config.legendSettings != default_legendSettings_asDict
            ):
                return None
            if (
                self.config.configOptionsExpanded != default_configOptionsExpanded
                and self.config.configOptionsExpanded
                != default_configOptionsExpanded_asDict
            ):
                return None
            if len(self.config.series) != 1:
                return None

            # At this point we can turn the series into the params
            series = self.config.series[0]
            mark = series.constants.mark
            if mark != None:
                field_vals.append(("mark", codify.object_to_code_no_format(mark)))
            for dim_name in [
                "color",
                "label",
                "pointShape",
                "tooltip",
                "x",
                "y",
                "y2",
            ]:
                table_col_id = getattr(series.dims, dim_name)
                sel_fn = series.table.columnSelectFunctions[table_col_id]
                if not isinstance(sel_fn, graph.VoidNode):
                    field_vals.append(
                        (
                            dim_name,
                            codify.lambda_wrapped_object_to_code_no_format(
                                sel_fn, ["row"]
                            ),
                        )
                    )

        param_str = ""
        if len(field_vals) > 0:
            param_str = (
                ",".join([f_name + "=" + f_val for f_name, f_val in field_vals]) + ","
            )
        return f"""weave.panels.panel_plot.Plot({codify.object_to_code_no_format(self.input_node)}, {param_str})"""


def make_set_all_series(dim_name: str) -> typing.Callable[[Plot, typing.Any], None]:
    def set(self: Plot, expr) -> None:
        if self.config is None:
            raise errors.WeaveInternalError("config is None")
        for series in self.config.series:
            setter = getattr(series, f"set_{dim_name}")
            setter(expr)

    return set


def make_group_all_series(dim_name: str) -> typing.Callable[[Plot], None]:
    def group(self: Plot) -> None:
        if self.config is None:
            raise errors.WeaveInternalError("config is None")
        for series in self.config.series:
            grouper = getattr(series, f"groupby_{dim_name}")
            grouper()

    return group


def make_get_all_series(
    dim_name: str,
) -> typing.Callable[[Plot], typing.List[graph.Node]]:
    def get(self: Plot) -> typing.List[graph.Node]:
        if self.config is None:
            raise errors.WeaveInternalError("config is None")
        dim_select_functions: typing.List[graph.Node] = []
        for series in self.config.series:
            getter = getattr(series, f"get_{dim_name}")
            dim_select_functions.append(getter())
        return dim_select_functions

    return get


def make_set_constant_all_series(
    const_name: str,
) -> typing.Callable[[Plot, typing.Any], None]:
    def set(self: Plot, value) -> None:
        if self.config is None:
            raise errors.WeaveInternalError("config is None")
        for series in self.config.series:
            setter = getattr(series, f"set_{const_name}_constant")
            setter(value)

    return set


def make_get_constant_all_series(
    const_name: str,
) -> typing.Callable[[Plot], typing.List[typing.Any]]:
    constants: typing.List[typing.Any] = []

    def get(self: Plot) -> typing.List[typing.Any]:
        if self.config is None:
            raise errors.WeaveInternalError("config is None")
        for series in self.config.series:
            getter = getattr(series, f"get_{const_name}_constant")
            constants.append(getter())
        return constants

    return get


# install setters and groupers
for dim in dataclasses.fields(DimConfig):
    setattr(Plot, f"set_{dim.name}", make_set_all_series(dim.name))
    setattr(Plot, f"groupby_{dim.name}", make_group_all_series(dim.name))
    setattr(Plot, f"get_{dim.name}", make_get_all_series(dim.name))

for dim in dataclasses.fields(PlotConstants):
    setattr(Plot, f"set_{dim.name}_constant", make_set_constant_all_series(dim.name))
    setattr(Plot, f"get_{dim.name}_constant", make_get_constant_all_series(dim.name))


def selection_is_continuous(selection: Selection) -> bool:
    return (
        isinstance(selection, (tuple, list))
        and len(selection) == 2
        and all(isinstance(i, (int, float)) for i in selection)
    )


def selection_is_discrete(selection: Selection) -> bool:
    return not selection_is_continuous(selection)


# TODO: keep in arrow
def filter_node_to_selection(
    node: graph.Node, selection: LazySelection, key: str
) -> graph.Node:
    if isinstance(selection, graph.Node):
        selection = weave_internal.call_fn(selection, {})

    selection = typing.cast(Selection, selection)

    if selection_is_continuous(selection):
        selection = typing.cast(ContinuousSelection, selection)

        def predicate(row: graph.Node[types.TypedDict]):
            target = dict_.TypedDict.pick(row, key)

            selection_min = selection[0]  # type: ignore
            selection_max = selection[1]  # type: ignore

            # This is a hack to handle selection of data from plots that have a temporal
            # range defined with bins ("start", "stop") instead of a single timestamp.
            # Should naievly handle selection of data from plots using op timestamps-binsnice
            if (
                weave.types.TypedDict({"start": weave.types.Timestamp()}).assign_type(
                    target.type
                )
                and isinstance(selection_min, (int, float))
                and isinstance(selection_max, (int, float))
            ):
                target = target["start"].toNumber()
            elif (
                weave.types.Timestamp().assign_type(target.type)
                and isinstance(selection_min, (int, float))
                and isinstance(selection_max, (int, float))
            ):
                target = target.toNumber()

            return boolean.Boolean.bool_and(
                target >= selection_min, target <= selection_max
            )

    else:

        def predicate(row: graph.Node[types.TypedDict]):
            target = dict_.TypedDict.pick(row, key)
            return list_.contains(selection, target)

    node_type = typing.cast(
        types.List,
        node.type,
    )

    filter = weave_internal.define_fn({"row": node_type.object_type}, predicate)
    return list_.List.filter(node, filter)


# TODO: keep in arrow
def _get_selected_rows_node(plot: Plot) -> graph.Node:
    if plot.config is None:
        raise errors.WeaveInternalError("config is None")

    selection = plot.config.signals.selection

    table_nodes: typing.List[graph.Node[types.List]] = []
    for series in plot.config.series:
        table_config = panel_table.TableConfig(tableState=series.table)
        table_panel = panel_table.Table(plot.input_node, config=table_config)
        node = list_.unnest(panel_table.rows(table_panel))
        columns = table_panel.get_final_named_select_functions()

        if selection.x is None and selection.y is None:
            # nothing selected
            table_nodes.append(list_.make_list())
            continue

        if selection.x is not None:
            col_id = series.dims.x
            col = columns.get(col_id, None)
            if col is not None:
                key = col["columnName"]
                node = filter_node_to_selection(node, selection.x, key)

        if selection.y is not None:
            col_id = series.dims.y
            col = columns.get(col_id, None)
            if col is not None:
                key = col["columnName"]
                node = filter_node_to_selection(node, selection.y, key)

        if len(plot.config.series) > 1:

            def merge_func(row: graph.Node[types.TypedDict]):
                return dict_.TypedDict.merge(
                    row, dict_.dict_(y_expr_str=series.y_expr_str)
                )

            func = weave_internal.define_fn({"row": node.type.object_type}, merge_func)
            node = list_.List.map(node, func)

        table_nodes.append(node)

    node_list = list_.make_list(
        **{str(i): node for (i, node) in enumerate(table_nodes)}
    )
    concatted = list_.List.concat(node_list)
    return concatted


# TODO: keep in arrow
def _get_selected_data_node(plot: Plot) -> graph.Node:
    if plot.config is None:
        raise errors.WeaveInternalError("config is None")

    selection = plot.config.signals.selection

    table_nodes: typing.List[graph.Node[types.List]] = []
    for series in plot.config.series:
        table_config = panel_table.TableConfig(tableState=series.table)
        table_config.tableState.add_column(lambda row: row, "row")
        table_panel = panel_table.Table(plot.input_node, config=table_config)
        node = list_.unnest(panel_table._get_rows_node(table_panel))
        columns = table_panel.get_final_named_select_functions()

        if selection.x is None and selection.y is None:
            # nothing selected
            table_nodes.append(list_.make_list())
            continue

        if selection.x is not None:
            col_id = series.dims.x
            col = columns.get(col_id, None)
            if col is not None:
                key = col["columnName"]
                node = filter_node_to_selection(node, selection.x, key)

        if selection.y is not None:
            col_id = series.dims.y
            col = columns.get(col_id, None)
            if col is not None:
                key = col["columnName"]
                node = filter_node_to_selection(node, selection.y, key)

        if len(plot.config.series) > 1:

            def merge_func(row: graph.Node[types.TypedDict]):
                return dict_.TypedDict.merge(
                    row, dict_.dict_(y_expr_str=series.y_expr_str)
                )

            func = weave_internal.define_fn({"row": node.type.object_type}, merge_func)
            node = list_.List.map(node, func)

        table_nodes.append(node)

    node_list = list_.make_list(
        **{str(i): node for (i, node) in enumerate(table_nodes)}
    )
    concatted = list_.List.concat(node_list)
    if concatted.type.object_type == types.UnknownType():
        return concatted
    return concatted["row"]


# TODO: keep in arrow
@weave.op(name="panel_plot-selected_rows_refine", hidden=True)
def selected_rows_refine(self: Plot) -> weave.types.Type:
    if self.config is None:
        raise errors.WeaveInternalError("config is None")

    nodes = []
    for series in self.config.series:
        table_config = panel_table.TableConfig(tableState=series.table)
        table_panel = panel_table.Table(self.input_node, config=table_config)
        node = list_.unnest(panel_table.rows(table_panel))
        if len(self.config.series) > 1:
            node = dict_.TypedDict.merge(
                node[0], dict_.dict_(y_expr_str=series.y_expr_str)
            )
        else:
            node = node[0]
        nodes.append(node)
    table_row_types = [node.type for node in nodes]  # type: ignore
    return types.List(types.union(*table_row_types))


# TODO: keep in arrow
@weave.op(name="panel_plot-selected_data_refine", hidden=True)
def selected_data_refine(self: Plot) -> weave.types.Type:
    return types.List(typing.cast(weave.types.List, self.input_node.type).object_type)


# TODO: keep in arrow
@weave.op(
    name="panel_plot-selected_rows",
    output_type=weave.types.List(weave.types.TypedDict({})),
    refine_output_type=selected_rows_refine,
)
def selected_rows(self: Plot):
    selected_rows_node = _get_selected_rows_node(self)
    return weave.use(selected_rows_node)


# TODO: keep in arrow
@weave.op(
    name="panel_plot-selected_data",
    output_type=weave.types.List(weave.types.Any()),
    refine_output_type=selected_data_refine,
)
def selected_data(self: Plot):
    selected_data_node = _get_selected_data_node(self)
    return selected_data_node



import contextlib
import contextvars
import typing
import copy

if typing.TYPE_CHECKING:
    from .run import Run

_run_stack: contextvars.ContextVar[list["Run"]] = contextvars.ContextVar(
    "run", default=[]
)


@contextlib.contextmanager
def current_run(
    run: "Run",
) -> typing.Iterator[list["Run"]]:
    new_stack = copy.copy(_run_stack.get())
    new_stack.append(run)

    token = _run_stack.set(new_stack)
    try:
        yield new_stack
    finally:
        _run_stack.reset(token)


def get_current_run() -> typing.Optional["Run"]:
    return _run_stack.get()[-1] if _run_stack.get() else None


def get_run_stack() -> list["Run"]:
    return _run_stack.get()


@contextlib.contextmanager
def set_run_stack(
    stack: list["Run"],
) -> typing.Iterator[list["Run"]]:
    token = _run_stack.set(stack)
    try:
        yield stack
    finally:
        _run_stack.reset(token)



import sys

# We track what modules were loaded before importing weave, so we can ensure
# that someone doesn't introduce auto-importing loading weave.ops or weave.panels
# (because they are slow to import and have more dependencies, and they are part of
# the engine and UI layers which should be kept separate from the core layer).
pre_init_modules = set(sys.modules.keys())

from . import context_state as _context_state

_loading_builtins_token = _context_state.set_loading_built_ins()

from . import weave_types as types
from . import storage

from .api import *


from .errors import *

from . import mappers_python_def

from . import wandb_api as _wandb_api

from . import version

_wandb_api.init()

# Ensure there is a client available for eager mode
from . import context as _context

_context_state.clear_loading_built_ins(_loading_builtins_token)

__version__ = version.VERSION


# See the comment above pre_init_modules above. This is check to ensure we don't accidentally
# introduce loading weave.ops or weave.panels when importing weave.
newly_added_modules = set(sys.modules.keys()) - pre_init_modules
ops_modules = []
panels_modules = []
for module_name in newly_added_modules:
    if module_name.startswith("weave.ops"):
        ops_modules.append(module_name)
    if module_name.startswith("weave.panels"):
        panels_modules.append(module_name)
if ops_modules or panels_modules:
    all_invalid_modules = ops_modules + panels_modules
    invalid_submodules = set([".".join(m.split(".")[:2]) for m in all_invalid_modules])
    raise errors.WeaveInternalError(
        "importing weave should not import weave.ops or weave.panels, but the following modules were imported: "
        + ", ".join(invalid_submodules)
    )



import typing

import weave
from .chat_model import ChatModel


@weave.type()
class StructuredOutputChatModel:
    @weave.op()
    def complete(
        self, messages: list[typing.Any], output_type: weave.types.Type
    ) -> typing.Any:
        ...


@weave.type()
class StructuredOutputChatModelSystemPrompt(StructuredOutputChatModel):
    chat_llm: ChatModel

    @weave.op()
    def complete(
        self, messages: list[typing.Any], output_type: weave.types.Type
    ) -> typing.Any:
        import json

        mess = messages + [
            {
                "role": "system",
                "content": f"Your response should only include a json object that matches the following type definition: {str(output_type)}",
            }
        ]
        response = self.chat_llm.complete(mess)
        parsed_response = json.loads(response["choices"][0]["message"]["content"])
        if not output_type.assign_type(weave.type_of(parsed_response)):
            raise ValueError("invalid response")
        return parsed_response



import typing


from . import weave_types as types
from . import errors
from . import ops
from . import ops_domain


class Weave0TypeJson(typing.TypedDict):
    wb_type: str
    params: typing.Any


def weave0_type_json_to_weave1_type(old_type_json: Weave0TypeJson) -> types.Type:
    return _convert_type(old_type_json)


primary_key_type_names = set(["primaryKey", "wandb.TablePrimaryKey"])
foreign_key_type_names = set(["foreignKey", "wandb.TableForeignKey"])
foreign_index_type_names = set(["foreignIndex", "wandb.TableForeignIndex"])


# This should ideally match `mediaTable.ts` in Weave0.
def _convert_type(old_type: Weave0TypeJson) -> types.Type:
    old_type_name = old_type["wb_type"]
    is_python_object_type = old_type_name == "pythonObject" or old_type_name == "object"
    #
    # Type Type
    #
    # TODO is this right?
    if old_type_name == "":
        return types.TypeType()
    #
    # General Types
    #
    elif old_type_name == "any":
        return types.Any()
    elif old_type_name == "unknown":
        return types.UnknownType()
    elif old_type_name == "invalid":
        return types.Invalid()
    #
    # Primitive Types
    #
    elif old_type_name == "none":
        return types.none_type
    elif old_type_name == "boolean":
        return types.Boolean()
    elif old_type_name == "number":
        # TODO: should we try to figure out if its an Int?
        return types.Number()
    elif old_type_name == "string":
        return types.String()
    #
    # Special Types
    #
    elif old_type_name == "union":
        return types.union(
            *(
                _convert_type(old_type)
                for old_type in old_type["params"]["allowed_types"]
            )
        )
    elif old_type_name == "const":
        val = old_type["params"]["val"]
        # if old_type["params"]["is_set"]:
        # TODO is_set is wrong... should be enum probably
        # val = set(val)
        return types.Const(types.TypeRegistry.type_of(val), val)
    #
    # Container Types
    #
    elif old_type_name == "typedDict" or old_type_name == "dictionary":
        return types.TypedDict(
            {
                prop_name: _convert_type(old_type)
                for prop_name, old_type in old_type["params"]["type_map"].items()
            }
        )
    elif old_type_name == "list":
        return types.List(_convert_type(old_type["params"]["element_type"]))
    #
    # Domain Types
    #
    elif old_type_name == "timestamp":
        return types.Timestamp()
    # Once Weave1 is fully launched and we don't have shadow mode, we can
    # uncomment the following block. Unfortunately, Weave0 views this as
    # unknown, and in shadow mode we don't use Weave1's type system.
    # elif (
    #     old_type_name == "pythonObject"
    #     and old_type.get("params", {}).get("class_name") == "Timestamp"
    # ):
    #     return types.Timestamp()
    elif (
        old_type_name == "pythonObject"
        and old_type.get("params", {}).get("class_name") == "datetime64"
    ):
        # This is a bit unfortunate. Weave0 interprets this as a number, then calls
        # something like `number-toTimestamp` on it. In the future, this should return
        # timestamp, but for now we'll just return number. Once we convert to Weave1,
        # we can change this because the frontend will be using this type instead of
        # figuring it out itself.
        return types.Int()
    elif old_type_name == "ndarray":
        # return ops.LegacyTableNDArrayType()
        if old_type.get("params", {}).get("serialization_path") is not None:
            # Just return NoneType. The data is None.
            return types.NoneType()

        # SDK converted the array to a pylist
        return types.List(types.UnknownType())

        # return NumpyArrayType("f", shape=old_type._params.get("shape", (0,)))
    #
    # Media Types
    #
    elif (
        old_type_name == "image-file"
        or old_type_name == "wandb.Image"
        or (is_python_object_type and old_type["params"].get("class_name") == "Image")
    ):
        if "params" not in old_type or "class_map" not in old_type["params"]:
            # This is legacy and fixed in `_patch_legacy_image_file_types``
            return ops.LegacyImageArtifactFileRefType()  # type: ignore

        boxLayersType = (
            weave0_type_json_to_weave1_type(old_type["params"]["box_layers"]).val  # type: ignore
            if "box_layers" in old_type["params"]
            else None
        )
        boxScoreKeysType = (
            weave0_type_json_to_weave1_type(old_type["params"]["box_score_keys"]).val  # type: ignore
            if "box_score_keys" in old_type["params"]
            else None
        )
        maskLayersType = (
            weave0_type_json_to_weave1_type(old_type["params"]["mask_layers"]).val  # type: ignore
            if "mask_layers" in old_type["params"]
            else None
        )
        classMapType = (
            weave0_type_json_to_weave1_type(old_type["params"]["class_map"]).val  # type: ignore
            if "class_map" in old_type["params"]
            else None
        )
        return ops.ImageArtifactFileRefType(  # type: ignore
            boxLayersType,
            boxScoreKeysType,
            maskLayersType,
            classMapType,
        )
    elif old_type_name == "audio-file" or (
        is_python_object_type and old_type["params"].get("class_name") == "Audio"
    ):
        return ops.AudioArtifactFileRef.WeaveType()  # type: ignore
    elif old_type_name == "html-file" or (
        is_python_object_type and old_type["params"].get("class_name") == "Html"
    ):
        return ops.HtmlArtifactFileRef.WeaveType()  # type: ignore
    elif old_type_name == "bokeh-file" or (
        is_python_object_type and old_type["params"].get("class_name") == "Bokeh"
    ):
        return ops.BokehArtifactFileRef.WeaveType()  # type: ignore
    elif old_type_name == "video-file" or (
        is_python_object_type and old_type["params"].get("class_name") == "Video"
    ):
        return ops.VideoArtifactFileRef.WeaveType()  # type: ignore
    elif old_type_name == "object3D-file" or (
        is_python_object_type and old_type["params"].get("class_name") == "Object3D"
    ):
        return ops.Object3DArtifactFileRef.WeaveType()  # type: ignore
    elif old_type_name == "classesId" or old_type_name == "wandb.Classes_id":
        # Question: Should we be converting this to a ClassesId type? Maybe a union of const numbers?
        # TODO: yes we should! It can be a pointer to a class set!
        return types.Number()  # type: ignore
    elif is_python_object_type and old_type["params"].get("class_name") == "Molecule":
        return ops.MoleculeArtifactFileRef.WeaveType()  # type: ignore

    elif old_type_name == "wb_trace_tree":
        return ops_domain.trace_tree.WBTraceTree.WeaveType()  # type: ignore

    #
    # Table Types
    #
    elif old_type_name == "table" or old_type_name == "wandb.Table":
        # TODO: this is not implemented in mediaTable.ts, leaving it as a TODO for now
        pass
    elif old_type_name == "joined-table":
        # TODO: this is not implemented in mediaTable.ts, leaving it as a TODO for now
        pass
    elif old_type_name == "partitioned-table":
        # TODO: this is not implemented in mediaTable.ts, leaving it as a TODO for now
        pass
    elif old_type_name in primary_key_type_names:
        return types.String()
    elif old_type_name in foreign_key_type_names:
        return types.String()
    elif old_type_name in foreign_index_type_names:
        return types.Number()
    #
    # Legacy Fallback Types
    #
    elif is_python_object_type:
        return types.UnknownType()
    raise errors.WeaveInternalError(
        "converting old Weave type not yet implemented: %s" % type(old_type)
    )



import dataclasses
import weave
import typing
from .. import panel
from .. import panel_util
from .. import ops
from .. import graph
from .. import weave_internal


@weave.type()
class WeaveLinkConfig:
    to: weave.Node[typing.Optional[typing.Any]] = dataclasses.field(
        default_factory=lambda: weave.graph.VoidNode()
    )
    vars: typing.Dict[str, graph.Node] = dataclasses.field(default_factory=dict)


@weave.type()
class WeaveLink(panel.Panel):
    id = "weavelink"
    input_node: graph.Node[typing.Any]

    def __init__(
        self,
        input_node,
        # This is frame
        to: typing.Callable[
            [graph.VarNode, typing.Dict[str, graph.VarNode]], graph.Node
        ],
        vars: typing.Dict[str, graph.Node] = {},
        config=None,
    ):
        super().__init__(input_node=input_node)
        self.config = config
        if self.config is None:
            self.config = WeaveLinkConfig(vars=vars)
            self.set_to(to)

    def set_to(self, to_node):
        input_node = self.input_node
        if self.config.vars:
            self.config.to = to_node(
                input_node,
                {
                    var_name: weave_internal.make_var_node(node.type, var_name)
                    for var_name, node in self.config.vars.items()
                },
            )
        else:
            self.config.to = to_node(input_node)

    # @property
    # def config(self):
    #     return {
    #         "to": self._to.to_json(),
    #         "vars": {var_name: node.to_json() for var_name, node in self._vars.items()},
    #     }



import typing
import weave


@weave.type()
class Dataset:
    rows: list[typing.Any]

    def __post_init__(self) -> None:
        if not isinstance(self.rows, weave.WeaveList):
            self.__dict__["rows"] = weave.WeaveList(self.rows)
